# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from concurrent.futures import Future

import pytest

from ansys.fluent.core.launcher.slurm_launcher import SlurmFuture


class SlurmEnvironment:
    def __init__(self):
        self.state = None

    def set_state(self, state):
        self.state = state


class SlurmFutureResult:
    pass


class SlurmFutureException:
    pass


@pytest.fixture
def slurm_future(monkeypatch: pytest.MonkeyPatch) -> SlurmFuture:
    env = SlurmEnvironment()
    future = Future()
    future.set_running_or_notify_cancel()
    slurm_future = SlurmFuture(future, 0)
    monkeypatch.setattr(slurm_future, "_get_state", lambda: env.state)
    monkeypatch.setattr(slurm_future, "_cancel", lambda: env.set_state(""))
    env.set_state("RUNNING")
    slurm_future.env = env
    return slurm_future


def test_cancel_slurm_future(slurm_future: SlurmFuture):
    assert slurm_future.cancel()


def test_slurm_future_lifecycle(slurm_future: SlurmFuture):
    assert slurm_future.pending()
    assert not slurm_future.running()
    assert not slurm_future.done()
    slurm_future._future.set_result(SlurmFutureResult())
    assert not slurm_future.pending()
    assert slurm_future.running()
    assert not slurm_future.done()
    assert isinstance(slurm_future.result(), SlurmFutureResult)
    slurm_future.env.set_state("")
    assert not slurm_future.pending()
    assert not slurm_future.running()
    assert slurm_future.done()


def test_slurm_future_exception(slurm_future: SlurmFuture):
    slurm_future._future.set_exception(SlurmFutureException())
    assert isinstance(slurm_future.exception(), SlurmFutureException)


def test_slurm_future_done_callback(slurm_future: SlurmFuture):
    called = []
    slurm_future.add_done_callback(lambda session: called.append(True))
    assert not called
    slurm_future._future.set_result(SlurmFutureResult())
    assert called
    assert called[0]



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Miscellaneous utility functions."""

import importlib.util
import logging
from pathlib import Path
import sys

from ansys.fluent.core.search import search  # noqa: F401

logger = logging.getLogger("pyfluent.general")


def load_module(module_name, file_path):
    """Load a module from a file path."""
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    logger.info(f"Loaded module {module_name} from {file_path}")
    return module


def get_examples_download_dir():
    """Return the path to the examples download directory."""
    parent_path = Path.home() / "Downloads"
    parent_path.mkdir(exist_ok=True)
    return parent_path / "ansys_fluent_core_examples"


def get_user_data_dir():
    """Return the path to the user data directory."""
    if sys.platform == "win32":
        return Path.home() / "AppData" / "Local" / "Ansys" / "ansys_fluent_core"
    else:
        return Path.home() / ".local" / "share" / "Ansys" / "ansys_fluent_core"



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for accessing and modifying hierarchy of Fluent settings.

The only useful method is '`get_root``, which returns the root object for
accessing Fluent settings.

Child objects can be generally accessed or modified using attribute access.
Named child objects can be accessed or modified using index operators.

Calling an object will return its current value.

Example
-------
>>> r = flobject.get_root(proxy)
>>> is_energy_on = r.setup.models.energy.enabled()
>>> r.setup.models.energy.enabled = True
>>> r.boundary_conditions.velocity_inlet['inlet'].vmag.constant = 20
"""

from __future__ import annotations

import collections
from contextlib import contextmanager, nullcontext
import fnmatch
import hashlib
import keyword
import logging
import os
import os.path
import pickle
import string
import sys
import types
from typing import (
    Any,
    Dict,
    ForwardRef,
    Generic,
    List,
    NewType,
    Tuple,
    TypeVar,
    Union,
    _eval_type,
    get_args,
    get_origin,
)
import warnings
import weakref

from ansys.fluent.core.pyfluent_warnings import (
    PyFluentDeprecationWarning,
    PyFluentUserWarning,
)
from ansys.fluent.core.utils.fluent_version import FluentVersion

from . import _docstrings
from .error_message import allowed_name_error_message, allowed_values_error
from .flunits import UnhandledQuantity, get_si_unit_for_fluent_quantity
from .settings_external import expand_api_file_argument


def _ansys_units():

    try:
        import ansys.units

        return ansys.units
    except ImportError:
        pass


settings_logger = logging.getLogger("pyfluent.settings_api")


class InactiveObjectError(RuntimeError):
    """Inactive object access."""

    def __init__(self, python_path):
        """Initialize InactiveObjectError."""
        super().__init__(f"'{python_path}' is currently inactive.")


class _InlineConstants:
    is_active = "active?"
    is_stable = "webui-release-active?"
    is_read_only = "read-only?"
    default_value = "default"
    min = "min"
    max = "max"
    user_creatable = "user-creatable?"
    allowed_values = "allowed-values"
    file_purpose = "file-purpose"


# Type hints
RealType = NewType("real", Union[float, str])  # constant or expression
RealListType = List[RealType]
RealVectorType = Tuple[RealType, RealType, RealType]
IntListType = List[int]
StringListType = List[str]
BoolListType = List[bool]
PrimitiveStateType = Union[
    str,
    RealType,
    int,
    bool,
    RealListType,
    IntListType,
    StringListType,
    BoolListType,
]
DictStateType = Dict[str, "StateType"]
ListStateType = List["StateType"]
StateType = Union[PrimitiveStateType, DictStateType, ListStateType]


def check_type(val, tp):
    """Check type of object."""
    if hasattr(tp, "__supertype__"):
        return check_type(val, tp.__supertype__)
    if isinstance(tp, ForwardRef):
        return check_type(val, _eval_type(tp, globals(), locals()))
    origin = get_origin(tp)
    if origin == list:
        return isinstance(val, list) and all(
            check_type(x, get_args(tp)[0]) for x in val
        )
    elif origin == tuple:
        return isinstance(val, tuple) and all(
            check_type(x, t) for x, t in zip(val, get_args(tp))
        )
    elif origin == Union:
        return any(check_type(val, t) for t in get_args(tp))
    elif origin == dict:
        k_t, k_v = get_args(tp)
        return isinstance(val, dict) and all(
            check_type(k, k_t) and check_type(v, k_v) for k, v in val.items()
        )
    elif origin is None:
        try:
            return isinstance(val, tp)
        except TypeError:
            return False
    else:
        return False


def assert_type(val, tp):
    """Assert type.

    Raises
    ------
    TypeError
        If the given value is not of the given type.
    """
    if not check_type(val, tp):
        raise TypeError(f"{val} is not of type {tp}.")


_ttable = str.maketrans(string.punctuation, "_" * len(string.punctuation), "?'")


def to_python_name(fluent_name: str) -> str:
    """Convert a scheme string to a Python variable name.

    This function replaces symbols with _. Any ``?`` symbols are
    ignored.
    """
    if not fluent_name:
        return fluent_name
    name = fluent_name.translate(_ttable)
    while name in keyword.kwlist:
        name = name + "_"
    return name


def _get_python_path_comps(obj):
    """Get python path components for traversing class hierarchy."""
    comps = []
    while obj:
        python_name = obj.python_name
        obj = obj._parent
        if isinstance(obj, (NamedObject, ListObject)):
            comps.append(obj.python_name)
            obj = obj._parent
        else:
            comps.append(python_name)
    comps.reverse()
    return comps[1:]


def _get_class_from_paths(root_cls, some_path: list[str], other_path: list[str]):
    """Get the class for the given alias path."""
    parent_count = 0
    while other_path[0] == "..":
        parent_count += 1
        other_path.pop(0)
    for _ in range(parent_count):
        some_path.pop()
    full_path = some_path + other_path
    cls = root_cls
    for comp in full_path:
        cls = cls._child_classes[comp]
        if issubclass(cls, (NamedObject, ListObject)):
            cls = cls.child_object_type
    return cls, full_path


class Base:
    """Provides the base class for settings and command objects.

    Parameters
    ----------
    name : str
        Name of the object if a child of a named object.
    parent: Base
           Parent of the object.

    Attributes
    ----------
    flproxy
    obj_name
    fluent_name
    """

    def __init__(self, name: str | None = None, parent=None):
        """__init__ of Base class."""
        self._setattr("_parent", weakref.proxy(parent) if parent is not None else None)
        self._setattr("_flproxy", None)
        self._setattr("_file_transfer_service", None)
        if name is not None:
            self._setattr("_name", name)
        self._setattr("_child_alias_objs", {})

    @property
    def _root(self):
        if self._parent is None:
            return self
        else:
            return self._parent._root

    def set_flproxy(self, flproxy):
        """Set flproxy object."""
        self._setattr("_flproxy", flproxy)

    def _set_on_interrupt(self, on_interrupt):
        """Set interrupt method."""
        self._setattr("_on_interrupt", on_interrupt)

    def _set_file_transfer_service(self, file_transfer_service):
        """Set file_transfer_service."""
        self._setattr("_file_transfer_service", file_transfer_service)

    @property
    def flproxy(self):
        """Proxy object.

        The proxy object is set at the root level and accessed via the parent for the
        child classes.
        """
        if self._flproxy is None:
            return self._parent.flproxy
        return self._flproxy

    @property
    def _file_transfer_handler(self):
        """Remote file handler.

        Supports file upload and download.
        """
        with warnings.catch_warnings():
            warnings.filterwarnings(action="ignore", category=UnstableSettingWarning)
            if self._file_transfer_service:
                return self._file_transfer_service
            elif self._parent:
                return self._parent._file_transfer_handler

    _name = None
    fluent_name = None
    _python_name = None

    @property
    def parent(self):
        """Parent (container) object."""
        return self._parent

    @property
    def obj_name(self) -> str:
        """Name of the scheme of this object.

        By default, this returns the object's static name. If the object is a child of a
        named object, the object's name is returned.
        """
        if self._name is None:
            return self.fluent_name
        return self._name

    @property
    def python_name(self) -> str:
        """Python name of this object.

        By default, this returns the object's static name. If the object is a child of a
        named object, the object's name is returned.
        """
        return getattr(self, "_python_name", None) or self.__class__.__name__

    @property
    def path(self) -> str:
        """Path of the object.

        Constructed from the ``obj_name`` of self and the path of
        parent.
        """
        if self._parent is None:
            return self.obj_name
        ppath = self._parent.path
        if not ppath:
            return self.obj_name
        return ppath + "/" + self.obj_name

    @property
    def python_path(self) -> str:
        """Path of the object.

        Constructed in python syntax from 'python_path' and the parents python path.
        """
        if self._parent is None:
            if FluentVersion(self._version).number < 251:
                return "<session>"
            else:
                return "<session>.settings"
        ppath = self._parent.python_path
        if not ppath:
            return self.python_name
        if self.python_name[0] == "[":
            return ppath + self.python_name
        return ppath + "." + self.python_name

    def get_attrs(self, attrs, recursive=False) -> Any:
        """Get the requested attributes for the object."""
        return self.flproxy.get_attrs(self.path, attrs, recursive)

    def get_attr(
        self,
        attr: str,
        attr_type_or_types: type | Tuple[type] | None = None,
    ) -> Any:
        """Get the requested attribute for the object.

        Parameters
        ----------
        attr : str
            attribute name
        attr_type_or_types : type or tuple of type, optional
            attribute type, by default None

        Returns
        -------
        Any
            attribute value

        Raises
        ------
        InactiveObjectError
            If any attribute other than ``"active?`` is queried when the object is not active.
        """
        attrs = self.get_attrs([attr])
        if attrs:
            attrs = attrs.get("attrs", attrs)
        if attr != "active?" and attrs and attrs.get("active?", True) is False:
            raise InactiveObjectError(self.python_path)
        val = None
        if attrs:
            val = attrs[attr]

        if attr_type_or_types:
            if not isinstance(attr_type_or_types, tuple):
                attr_type_or_types = (attr_type_or_types,)
            if isinstance(val, attr_type_or_types):
                return val
            if val is not None and any(
                issubclass(x, bool) for x in attr_type_or_types
            ):  # cast to bool for boolean attributes
                return bool(val)
            return None
        return val

    def _is_deprecated(self) -> bool:
        """Whether the object is deprecated in a specific Fluent version.'"""
        deprecated_version = self.get_attrs(["deprecated-version"])
        deprecated_version = (
            deprecated_version.get("deprecated-version") if deprecated_version else None
        )
        return deprecated_version and (
            float(deprecated_version) <= 22.2
            or FluentVersion(self._version) >= FluentVersion(deprecated_version)
        )

    def is_active(self) -> bool:
        """Whether the object is active."""
        attr = self.get_attr(_InlineConstants.is_active) and not self._is_deprecated()
        return False if attr is False else True

    def _check_stable(self) -> None:
        """Whether the object is stable."""
        if not self.is_active():
            return
        attr = self.get_attr(_InlineConstants.is_stable)
        attr = True if attr is None else attr
        if not attr:
            warnings.warn(
                f"The API feature at '{self.path}' is not stable. "
                f"It is not guaranteed that it is fully validated and "
                f"there is no commitment to its backwards compatibility.",
                UnstableSettingWarning,
            )

    def is_read_only(self) -> bool:
        """Whether the object is read-only."""
        attr = self.get_attr(_InlineConstants.is_read_only)
        return False if attr is None else attr

    def __setattr__(self, name, value):
        raise AttributeError(name)

    # __setattr__ is overridden to prevent creation of new attributes or
    # overriding existing ones. _setattr is the backdoor to set attributes
    def _setattr(self, name, value):
        super().__setattr__(name, value)

    def find_object(self, relative_path):
        """Find object."""
        obj = self
        for comp in relative_path.split("/"):
            if comp == "..":
                obj = obj.parent
            else:
                obj = getattr(obj, comp)
        return obj

    def before_execute(self, command_name, value, kwargs):
        """Executes before command execution."""
        if hasattr(self, "_do_before_execute"):
            base_file_name = self._do_before_execute(
                command_name=command_name, value=value, kwargs=kwargs
            )
            return base_file_name
        else:
            return value

    def after_execute(self, command_name, value, kwargs):
        """Executes after command execution."""
        if hasattr(self, "_do_after_execute"):
            base_file_name = self._do_after_execute(
                command_name=command_name, value=value, kwargs=kwargs
            )
            return base_file_name
        else:
            return value

    def _while_setting_state(self):
        """Avoid additional processing while setting the state."""
        return nullcontext()

    def _while_renaming(self):
        """Avoid additional processing while renaming."""
        return nullcontext()

    def _while_deleting(self):
        """Avoid additional processing while deleting."""
        return nullcontext()

    def _while_creating(self):
        """Avoid additional processing while creating."""
        return nullcontext()

    def _while_resizing(self):
        """Avoid additional processing while resizing."""
        return nullcontext()

    def _while_executing_command(self):
        """Avoid additional processing while executing a command."""
        return nullcontext()

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.flproxy == other.flproxy and self.path == other.path


StateT = TypeVar("StateT")


class Property(Base):
    """Exposes attribute accessor on settings object."""

    def default_value(self):
        """Gets the default value of the object."""
        return self.get_attr(_InlineConstants.default_value)


class Numerical(Property):
    """Exposes attribute accessor on settings object - specific to numerical objects."""

    def min(self):
        """Get the minimum value of the object."""
        val = self.get_attr(_InlineConstants.min, (float, int))
        return None if isinstance(val, bool) else val

    def max(self):
        """Get the maximum value of the object."""
        val = self.get_attr(_InlineConstants.max, (float, int))
        return None if isinstance(val, bool) else val


QuantityT = TypeVar("QuantityT")


class RealNumerical(Numerical):
    """A ``RealNumerical`` object representing a real value setting, including single
    real values and containers of real values, such as lists.

    Methods
    -------
    as_quantity()
        Get the current state of the object as an ansys.units.Quantity.

    set_state(state)
        Set the state of the object.

    units()
        Get the units string.
    """

    def as_quantity(self) -> QuantityT | None:
        """Get the state of the object as an ansys.units.Quantity."""
        error = None
        if not _ansys_units():
            error = "Code not configured to support units."
        if not error:
            quantity = self.get_attr("units-quantity")
            units = get_si_unit_for_fluent_quantity(quantity)
            if units is not None:
                try:
                    return _ansys_units().Quantity(
                        value=self.get_state(),
                        units=units,
                    )
                except (TypeError, ValueError) as e:
                    error = e
            else:
                error = "Could not determine units."
        warnings.warn(f"Unable to construct 'Quantity'. {error}")

    def set_state(self, state: StateT | None = None, **kwargs):
        """Set the state of the object.

        Parameters
        ----------
        state
            The type of state can be float, str (representing either
            an expression or a value with units), or an ansys.units.Quantity.
        kwargs : Any
            Keyword arguments.

        Raises
        ------
        UnhandledQuantity
            If the quantity object cannot be handled for the given path. This can
            happen if the quantity attribute specifies an unsupported quantity, or if
            the units specified for the quantity are not supported.
        """
        try:

            def get_units():
                units = self.units()
                if units is None:
                    raise UnhandledQuantity(self.path, state)
                return units

            if _ansys_units() and isinstance(state, (_ansys_units().Quantity, tuple)):
                state = (
                    _ansys_units().Quantity(*state)
                    if isinstance(state, tuple)
                    else state
                )
                state = state.to(get_units()).value
            elif isinstance(state, tuple):
                if state[1] == get_units():
                    state = state[0]
                else:
                    raise UnhandledQuantity(self.path, state)
        except Exception as ex:
            raise UnhandledQuantity(self.path, state) from ex

        return self.base_set_state(state=state, **kwargs)

    def units(self) -> str | None:
        """Get the physical units of the object as a string."""
        quantity = self.get_attr("units-quantity")
        return get_si_unit_for_fluent_quantity(quantity)


class Textual(Property):
    """Exposes attribute accessor on settings object - specific to string objects."""


class DeprecatedSettingWarning(PyFluentDeprecationWarning):
    """Provides deprecated settings warning."""

    pass


class UnstableSettingWarning(PyFluentUserWarning):
    """Provides unstable settings warning."""

    pass


_show_warning_orig = warnings.showwarning


def _show_warning(message, category, *args, **kwargs):
    if category == DeprecatedSettingWarning:
        print(message)
    else:
        _show_warning_orig(message, category, *args, **kwargs)


warnings.showwarning = _show_warning


class _Alias:
    scheme_eval = None
    once = False

    @contextmanager
    def _print_newer_api(self):
        scheme_eval = _Alias.scheme_eval
        if scheme_eval:
            scheme_eval("(define pyfluent-journal-str-port (open-output-string))")
            scheme_eval("(api-echo-python-port pyfluent-journal-str-port)")
        try:
            yield
        finally:
            if scheme_eval:
                scheme_eval("(api-unecho-python-port pyfluent-journal-str-port)")
                journal_str = scheme_eval(
                    "(close-output-port pyfluent-journal-str-port)"
                )
                if isinstance(journal_str, str):
                    warnings.warn(
                        "Note: A newer syntax is available to perform the last operation:\n"
                        f"{journal_str.strip()}",
                        DeprecatedSettingWarning,
                    )
                if not _Alias.once:
                    warnings.warn(
                        "\nExecute the following code to suppress future warnings like the above:\n\n"
                        ">>> import warnings\n"
                        '>>> warnings.filterwarnings("ignore", category=DeprecatedSettingWarning)',
                        DeprecatedSettingWarning,
                    )
                    _Alias.once = True

    def _while_setting_state(self):
        return self._print_newer_api()

    def _while_renaming(self):
        return self._print_newer_api()

    def _while_deleting(self):
        return self._print_newer_api()

    def _while_creating(self):
        return self._print_newer_api()

    def _while_resizing(self):
        return self._print_newer_api()

    def _while_executing_command(self):
        return self._print_newer_api()


def _create_child(cls, name, parent: weakref.CallableProxyType, alias_path=None):
    if alias_path or isinstance(parent, _Alias):
        bases = (cls,)
        # True child of parent alias is already derived from _Alias.
        # Avoiding duplicate derivation and mro resolution issue.
        if _Alias not in cls.__mro__:
            bases = (_Alias, cls)
        alias_cls = type(
            f"{cls.__name__}_alias",
            bases,
            dict(cls.__dict__) | {"alias_path": alias_path},
        )
        return alias_cls(name, parent.__repr__.__self__)
    return cls(name, parent)


class SettingsBase(Base, Generic[StateT]):
    """Base class for settings objects.

    Methods
    -------
    get_state()
        Get the current state of the object.

    set_state(state)
        Set the state of the object.
    """

    @classmethod
    def to_scheme_keys(cls, value: StateT, root_cls, path: list[str]) -> StateT:
        """Convert value to have keys with scheme names.

        This is overridden in the ``Group``, ``NamedObject``, and
        ``ListObject`` classes.
        """
        return value

    @classmethod
    def to_python_keys(cls, value: StateT) -> StateT:
        """Convert value to have keys with Python names.

        This is overridden in the ``Group``, ``NamedObject``, and
        ``ListObject`` classes.
        """
        return value

    def __call__(self) -> StateT:
        """Alias for self.get_state."""
        return self.get_state()

    def get_state(self) -> StateT:
        """Get the state of the object."""
        return self.to_python_keys(self.flproxy.get_var(self.path))

    def set_state(self, state: StateT | None = None, **kwargs):
        """Set the state of the object."""
        with self._while_setting_state():
            if isinstance(state, (tuple, _ansys_units().Quantity)) and hasattr(
                self, "value"
            ):
                self.value.set_state(state, **kwargs)
            else:
                self.flproxy.set_var(
                    self.path,
                    self.to_scheme_keys(
                        kwargs or state,
                        self._root.__class__,
                        _get_python_path_comps(self),
                    ),
                )

    @staticmethod
    def _print_state_helper(state, out, indent=0, indent_factor=2):
        if isinstance(state, dict):
            out.write("\n")
            for key, value in state.items():
                if value is not None:
                    out.write(f'{indent*indent_factor*" "}{key} : ')
                    SettingsBase._print_state_helper(
                        value, out, indent + 1, indent_factor
                    )
        elif isinstance(state, list):
            out.write("\n")
            for index, value in enumerate(state):
                out.write(f'{indent*indent_factor*" "}{index} : ')
                SettingsBase._print_state_helper(value, out, indent + 1, indent_factor)
        else:
            out.write(f"{state}\n")

    def print_state(self, out=None, indent_factor=2):
        """Print the state of the object."""
        out = sys.stdout if out is None else out
        self._print_state_helper(self.get_state(), out, indent_factor=indent_factor)

    def state_with_units(self) -> StateT:
        """Get the state of the object with units where available."""
        state = self.get_state()
        if isinstance(self, RealNumerical):
            return (state, self.units())
        elif isinstance(state, collections.abc.Mapping):
            self._add_units_to_state(state)
        return state

    def _add_units_to_state(self, state):
        if isinstance(state, collections.abc.Mapping):
            for k, v in state.items():
                child = None
                if isinstance(self, collections.abc.Mapping):
                    try:
                        child = self[k]
                    except KeyError:
                        pass
                child = child or getattr(self, k, None)
                if child is None:
                    raise RuntimeError(
                        "Unexpected None child {k} encountered while getting units for state."
                    )
                elif isinstance(child, RealNumerical):
                    state[k] = (state[k], child.units())
                else:
                    child._add_units_to_state(state[k])


class Integer(SettingsBase[int], Numerical):
    """An ``Integer`` object representing an integer value setting."""

    _state_type = int


class Real(SettingsBase[RealType], RealNumerical):
    """A ``Real`` object representing a real value setting.

    Some ``Real`` objects also accept string arguments representing
    expression values.
    """

    base_set_state = SettingsBase[RealType].set_state
    set_state = RealNumerical.set_state

    _state_type = RealType


class String(SettingsBase[str], Textual):
    """A ``String`` object representing a string value setting."""

    _state_type = str


class Filename(SettingsBase[str], Textual):
    """A ``Filename`` object representing a file name."""

    _state_type = str

    def file_purpose(self):
        """Specifies whether this file is used as input or output by Fluent."""
        return self.get_attr(_InlineConstants.file_purpose)


class FilenameList(SettingsBase[StringListType], Textual):
    """A FilenameList object represents a list of file names."""

    _state_type = StringListType

    def file_purpose(self):
        """Specifies whether this file is used as input or output by Fluent."""
        return self.get_attr(_InlineConstants.file_purpose)


class FileName(Base):
    """Resolves MRO for child classes."""

    pass


class _InputFile(FileName):
    def _do_before_execute(self, command_name, value, kwargs):
        file_names = expand_api_file_argument(command_name, value, kwargs)
        if self._file_transfer_handler:
            for file_name in file_names:
                self._file_transfer_handler.upload(file_name=file_name)
            return os.path.basename(value)
        else:
            return value


class _OutputFile(FileName):
    def _do_after_execute(self, command_name, value, kwargs):
        file_names = expand_api_file_argument(command_name, value, kwargs)
        if self._file_transfer_handler:
            for file_name in file_names:
                self._file_transfer_handler.download(file_name=file_name)
            return os.path.basename(value)
        else:
            return value


class _InOutFile(_InputFile, _OutputFile):
    pass


class Boolean(SettingsBase[bool], Property):
    """A ``Boolean`` object representing a Boolean value setting."""

    _state_type = bool


class RealList(SettingsBase[RealListType], RealNumerical):
    """A ``RealList`` object representing a real list setting."""

    base_set_state = SettingsBase[RealListType].set_state
    set_state = RealNumerical.set_state

    _state_type = RealListType


class IntegerList(SettingsBase[IntListType], Numerical):
    """An ``Integer`` object representing an integer list setting."""

    _state_type = IntListType


class RealVector(SettingsBase[RealVectorType], Numerical):
    """An object representing a 3D vector.

    A ``RealVector`` object representing a real vector setting
    consisting of three real values.
    """

    _state_type = RealVectorType


class StringList(SettingsBase[StringListType], Textual):
    """A ``StringList`` object representing a string list setting."""

    _state_type = StringListType


class BooleanList(SettingsBase[BoolListType], Property):
    """A ``BooleanList`` object representing a Boolean list setting."""

    _state_type = BoolListType


def _command_query_name_filter(
    parent, list_attr: str, prefix: str, excluded: List[str]
) -> List:
    """Auto completer info of commands and queries."""
    ret = []
    names = getattr(parent, list_attr)
    for name in names:
        if name not in excluded and name.startswith(prefix):
            child = getattr(parent, name)
            if child.is_active():
                ret.append([name, child.__class__.__bases__[0].__name__, child.__doc__])
    return ret


def _get_type_for_completer_info(cls) -> str:
    if issubclass(cls, (FileName, _InputFile)):
        return "InputFilename"
    elif issubclass(cls, (FileName, _OutputFile)):
        return "OutputFilename"
    elif issubclass(cls, (FileName, _InOutFile)):
        return "InOutFilename"
    elif issubclass(cls, (FilenameList, _InputFile)):
        return "InputFilenameList"
    elif issubclass(cls, (FilenameList, _OutputFile)):
        return "OutputFilenameList"
    elif issubclass(cls, (FilenameList, _InOutFile)):
        return "InOutFilenameList"
    else:
        return cls.__bases__[0].__name__


class Group(SettingsBase[DictStateType]):
    """A ``Group`` container object.

    A ``Group`` object is a container similar to a C++ structure object.
    Child objects can be accessed via attribute access.

    Attributes
    ----------
    child_names: list[str]
        Names of the child objects
    command_names: list[str]
        Names of the commands
    """

    _state_type = DictStateType

    def __init__(self, name: str | None = None, parent=None):
        """__init__ of Group class."""
        super().__init__(name, parent)
        for child in self.child_names:
            cls = self.__class__._child_classes[child]
            self._setattr(child, _create_child(cls, None, self))
        for cmd in self.command_names:
            cls = self.__class__._child_classes[cmd]
            self._setattr(cmd, _create_child(cls, None, self))
        for query in self.query_names:
            cls = self.__class__._child_classes[query]
            self._setattr(query, _create_child(cls, None, self))

    def __call__(self, *args, **kwargs):
        if kwargs:
            self.set_state(kwargs)
        elif args:
            self.set_state(args)
        else:
            return self.get_state()

    @classmethod
    def to_scheme_keys(cls, value, root_cls, path: list[str]):
        """Convert value to have keys with scheme names.

        Raises
        ------
        RuntimeError
            If key is invalid.
        """
        if isinstance(value, collections.abc.Mapping):
            ret = {}
            for k, v in value.items():
                if k in cls.child_names:
                    ccls = cls._child_classes[k]
                    ret[ccls.fluent_name] = ccls.to_scheme_keys(v, root_cls, path + [k])
                elif k in cls._child_aliases:
                    alias, scm_alias_name = cls._child_aliases[k]
                    alias_cls, alias_path = _get_class_from_paths(
                        root_cls, path.copy(), alias.split("/")
                    )
                    ret[scm_alias_name] = alias_cls.to_scheme_keys(
                        v, root_cls, alias_path
                    )
                else:
                    raise RuntimeError("Key '" + str(k) + "' is invalid")
            return ret
        else:
            return value

    @classmethod
    def to_python_keys(cls, value):
        """Convert value to have keys with Python names."""
        if isinstance(value, collections.abc.Mapping):
            ret = {}
            undef = object()
            for mname in cls.child_names:
                ccls = cls._child_classes[mname]
                mvalue = value.get(ccls.fluent_name, undef)
                if mvalue is not undef:
                    ret[mname] = ccls.to_python_keys(mvalue)
            return ret
        else:
            return {}

    _child_classes = {}
    child_names = []
    command_names = []
    query_names = []
    _child_aliases = {}

    def get_active_child_names(self):
        """Names of children that are currently active."""
        ret = []
        for child in self.child_names:
            if getattr(self, child).is_active():
                ret.append(child)
        return ret

    def get_active_command_names(self):
        """Names of commands that are currently active."""
        ret = []
        for command in self.command_names:
            if getattr(self, command).is_active():
                ret.append(command)
        return ret

    def get_active_query_names(self):
        """Names of queries that are currently active."""
        ret = []
        for query in self.query_names:
            if getattr(self, query).is_active():
                ret.append(query)
        return ret

    def __dir__(self):
        dir_list = set(list(self.__dict__.keys()) + dir(type(self)))
        return dir_list - set(
            [
                child
                for child in self.child_names + self.command_names + self.query_names
                if getattr(self, child)._is_deprecated()
            ]
        )

    def get_completer_info(self, prefix="", excluded=None) -> List[List[str]]:
        """Get completer info of all children.

        Returns
        -------
        List[List[str]]
            Name, type and docstring of all children.
        """
        excluded = excluded or []
        ret = []
        for child_name in self.child_names:
            if child_name not in excluded and child_name.startswith(prefix):
                child = getattr(self, child_name)
                if child.is_active():
                    ret.append(
                        [
                            child_name,
                            _get_type_for_completer_info(child.__class__),
                            child.__doc__,
                        ]
                    )
        command_info = _command_query_name_filter(
            self, "command_names", prefix, excluded
        )
        query_info = _command_query_name_filter(self, "query_names", prefix, excluded)
        for items in [command_info, query_info]:
            ret.extend(items)
        return ret

    def __getattribute__(self, name):
        if name in super().__getattribute__("child_names"):
            if self.is_active() is False:
                raise InactiveObjectError(self.python_path)
        alias = super().__getattribute__("_child_aliases").get(name)
        if alias:
            alias = alias[0]
            alias_obj = self._child_alias_objs.get(name)
            if alias_obj is None:
                obj = self.find_object(alias)
                alias_obj = self._child_alias_objs[name] = _create_child(
                    obj.__class__, None, obj.parent, alias
                )
            return alias_obj
        try:
            attr = super().__getattribute__(name)
            if name in super().__getattribute__("_child_classes"):
                attr._check_stable()
            return attr
        except AttributeError as ex:
            error_msg = allowed_name_error_message(
                trial_name=name,
                message=ex.args[0],
                allowed_values=sorted(
                    set(self.get_active_child_names() + self.command_names)
                ),
            )
            ex.args = (error_msg,)
            raise

    def __setattr__(self, name: str, value):
        attr = None
        try:
            attr = getattr(self, name)
        except AttributeError as ex:
            error_msg = allowed_name_error_message(
                trial_name=name,
                allowed_values=super().__getattribute__("child_names"),
                message=ex.args[0],
            )
            ex.args = (error_msg,)
            raise
        try:
            return attr.set_state(value)
        except Exception as ex:
            if hasattr(attr, "allowed_values"):
                allowed = attr.allowed_values()
                if allowed and value not in allowed:
                    raise allowed_values_error(name, value, allowed) from ex
            else:
                raise


class WildcardPath(Group):
    """Class wrapping a wildcard path to perform get_var and set_var on flproxy."""

    def __init__(self, flproxy, path: str, state_cls, settings_cls, parent):
        """__init__ of WildcardPath class."""
        self._setattr("_flproxy", flproxy)
        self._setattr("_path", path)
        # _state_cls is the settings class at which the state is constructed.
        # _state_cls isn't changed after the first wildcard, i.e.
        # a.b["*"], a.b["*"].c, a.b["*"].c.d["*"] have the same _state_cls.
        # It is used to convert between python and scheme keys within the state.
        self._setattr("_state_cls", state_cls)
        # _settings_cls is the settings cls at the wildcard path level. It is used to
        # construct the scheme path for children.
        self._setattr("_settings_cls", settings_cls)
        self._setattr("_parent", parent)

    @property
    def flproxy(self):
        """Proxy object."""
        return self._flproxy

    @property
    def path(self):
        """Path with wildcards."""
        return self._path

    def __getattr__(self, name: str):
        try:
            child_settings_cls = self._settings_cls._child_classes[name]
            scheme_name = child_settings_cls.fluent_name
            wildcard_cls = (
                NamedObjectWildcardPath
                if issubclass(child_settings_cls, NamedObject)
                else WildcardPath
            )
            return wildcard_cls(
                self.flproxy,
                self.path + "/" + scheme_name,
                self._state_cls,
                child_settings_cls,
                self,
            )
        except KeyError as ex:
            raise AttributeError(
                allowed_name_error_message(
                    context=self._state_cls._python_name,
                    trial_name=name,
                    allowed_values=self.get_active_child_names(),
                )
            ) from ex

    def items(self):
        """Items."""
        for key, value in self._parent.items():
            if fnmatch.fnmatch(key, self._path.rsplit(sep="/", maxsplit=1)[-1]):
                yield key, value

    def __iter__(self):
        for item in self._parent:
            if fnmatch.fnmatch(item, self._path.rsplit(sep="/", maxsplit=1)[-1]):
                yield item

    # Note that following 2 are not symmetric as get_state and set_state
    # are not symmetric.
    # get_state example: a.b["*"].c.d.get_state() == {"<bN>" {"c": {"d": <d_value>}}}
    # set_state example: a.b["*"].set_state({"c": {"d": <d_value>}})

    def to_scheme_keys(self, value, root_cls, path):
        """Convert value to have keys with scheme names."""
        return self._settings_cls.to_scheme_keys(value, root_cls, path)

    def to_python_keys(self, value):
        """Convert value to have keys with Python names."""
        return self._state_cls.to_python_keys(value)


class NamedObjectWildcardPath(WildcardPath):
    """WildcardPath at a NamedObject path, so it can be looked up by wildcard again."""

    def __getitem__(self, name: str):
        return WildcardPath(
            self.flproxy,
            self.path + "/" + name,
            self._state_cls,
            self._settings_cls.child_object_type,
            self,
        )

    def __setitem__(self, name, value):
        self[name].set_state(value)


ChildTypeT = TypeVar("ChildTypeT")


class NamedObject(SettingsBase[DictStateType], Generic[ChildTypeT]):
    """A ``NamedObject`` container is a container object similar to a Python dictionary
    object. Generally, many such objects can be created with different names.

    Attributes
    ----------
    command_names: list[str]
        Names of the commands
    """

    # New objects could get inserted by other operations, so we cannot assume
    # that the local cache in self._objects is always up-to-date
    def __init__(self, name: str | None = None, parent=None):
        """__init__ of NamedObject class."""
        super().__init__(name, parent)
        self._setattr("_objects", {})
        for cmd in self.command_names:
            cls = self.__class__._child_classes[cmd]
            self._setattr(cmd, _create_child(cls, None, self))
        for query in self.query_names:
            cls = self.__class__._child_classes[query]
            self._setattr(query, _create_child(cls, None, self))
        if not hasattr(
            self, "rename"
        ):  # if rename command is not available from settings API
            self._setattr(
                "rename",
                types.MethodType(lambda obj, new, old: _rename(obj, new, old), self),
            )

    @classmethod
    def to_scheme_keys(cls, value, root_cls, path: list[str]):
        """Convert value to have keys with scheme names."""
        if isinstance(value, collections.abc.Mapping):
            ret = {}
            for k, v in value.items():
                ret[k] = cls.child_object_type.to_scheme_keys(v, root_cls, path)
            return ret
        else:
            return value

    @classmethod
    def to_python_keys(cls, value):
        """Convert value to have keys with Python names."""
        if isinstance(value, collections.abc.Mapping):
            ret = {}
            for k, v in value.items():
                ret[k] = cls.child_object_type.to_python_keys(v)
            return ret
        else:
            return {}

    _child_classes = {}
    command_names = []
    query_names = []
    _child_aliases = {}

    def _create_child_object(self, cname: str):
        ret = self._objects.get(cname)
        if not ret:
            cls = self.__class__.child_object_type
            ret = self._objects[cname] = _create_child(cls, cname, self)
        ret._setattr("_python_name", f'["{cname}"]')
        ret._setattr(
            "rename",
            types.MethodType(lambda obj, name: _rename(self, name, cname), ret),
        )
        return ret

    def _update_objects(self):
        names = self.get_object_names()
        for name in list(self._objects.keys()):
            if name not in names:
                del self._objects[name]
        for name in names:
            if name not in self._objects:
                self._create_child_object(name)

    def __delitem__(self, name: str):
        with self._while_deleting():
            self.flproxy.delete(self.path, name)
        if name in self._objects:
            del self._objects[name]

    def __contains__(self, name: str):
        return name in self.get_object_names()

    def __len__(self):
        return len(self.keys())

    def __iter__(self):
        self._update_objects()
        return iter(self._objects)

    def keys(self):
        """Object names."""
        self._update_objects()
        return self._objects.keys()

    def values(self):
        """Object values."""
        self._update_objects()
        return self._objects.values()

    def items(self):
        """Items."""
        self._update_objects()
        return self._objects.items()

    def user_creatable(self) -> bool:
        """Whether the object is user-creatable."""
        return self.get_attr(_InlineConstants.user_creatable, bool)

    def get_object_names(self):
        """Object names."""
        obj_names = self.flproxy.get_object_names(self.path)
        obj_names_list = obj_names if isinstance(obj_names, list) else list(obj_names)
        return obj_names_list

    def get_completer_info(self, prefix="", excluded=None) -> List[List[str]]:
        """Get completer info of all children.

        Returns
        -------
        List[List[str]]
            Name, type and docstring of all children.
        """
        excluded = excluded or []
        ret = []
        command_info = _command_query_name_filter(
            self, "command_names", prefix, excluded
        )
        query_info = _command_query_name_filter(self, "query_names", prefix, excluded)
        for items in [command_info, query_info]:
            ret.extend(items)
        return ret

    def __getitem__(self, name: str) -> ChildTypeT:
        if name not in self.get_object_names():
            if self.flproxy.has_wildcard(name):
                child_cls = self.__class__.child_object_type
                # TODO: alias
                return WildcardPath(
                    self.flproxy,
                    self.path + "/" + name,
                    self.__class__,
                    child_cls,
                    self,
                )
            raise KeyError(
                allowed_name_error_message(
                    context=self.python_name,
                    trial_name=name,
                    allowed_values=self.get_object_names(),
                )
            )

        obj = self._objects.get(name)
        if not obj:
            obj = self._create_child_object(name)
        return obj

    def get(self, name: str) -> ChildTypeT:
        """Return the child object by key.

        Parameters
        ----------
        name : str
            Name of the child object.

        Returns
        -------
        ChildTypeT
            Child object.
        """
        try:
            return self.__getitem__(name)
        except Exception:
            return

    def __getattr__(self, name: str):
        alias = self._child_aliases.get(name)
        if alias:
            alias = alias[0]
            alias_obj = self._child_alias_objs.get(name)
            if alias_obj is None:
                obj = self.find_object(alias)
                alias_obj = self._child_alias_objs[name] = _create_child(
                    obj.__class__, None, obj.parent, alias
                )
            return alias_obj
        else:
            return getattr(super(), name)


def _rename(obj: NamedObject | _Alias, new: str, old: str):
    """Rename a named object.

    Parameters
    ----------
    obj: NamedObject
        named-object to be renamed
    new: str
        New name.
    old : str
        Current name.
    """
    with obj._while_renaming():
        obj.flproxy.rename(obj.path, new, old)
    if old in obj._objects:
        del obj._objects[old]
    obj._create_child_object(new)


class ListObject(SettingsBase[ListStateType], Generic[ChildTypeT]):
    """A ``ListObject`` container is a container object, similar to a Python list
    object. Generally, many such objects can be created.

    Attributes
    ----------
    command_names: list[str]
        Names of the commands.

    Methods
    -------
    get_size()
       Get the size of the list.
    """

    # New objects could get inserted by other operations, so we cannot assume
    # that the local cache in self._objects is always up-to-date
    def __init__(self, name=None, parent=None):
        """__init__ of ListObject class."""
        super().__init__(name, parent)
        self._setattr("_objects", [])
        for cmd in self.command_names:
            cls = self.__class__._child_classes[cmd]
            self._setattr(cmd, _create_child(cls, None, self))
        for query in self.query_names:
            cls = self.__class__._child_classes[query]
            self._setattr(query, _create_child(cls, None, self))

    @classmethod
    def to_scheme_keys(cls, value, root_cls, path: list[str]):
        """Convert value to have keys with scheme names."""
        if isinstance(value, collections.abc.Sequence):
            return [
                cls.child_object_type.to_scheme_keys(v, root_cls, path) for v in value
            ]
        else:
            return value

    @classmethod
    def to_python_keys(cls, value):
        """Convert value to have keys with scheme names."""
        if isinstance(value, collections.abc.Sequence):
            return [cls.child_object_type.to_python_keys(v) for v in value]
        else:
            return []

    _child_classes = {}
    command_names = []
    query_names = []
    _child_aliases = {}

    def _update_objects(self):
        cls = self.__class__.child_object_type
        self._setattr(
            "_objects",
            [_create_child(cls, str(x), self) for x in range(self.get_size())],
        )

    def __len__(self):
        return self.get_size()

    def __iter__(self):
        self._update_objects()
        return iter(self._objects)

    def get_size(self) -> int:
        """Return the number of elements in a list object.

        Returns
        -------
        int
        """
        return self.flproxy.get_list_size(self.path)

    def __getitem__(self, index: int) -> ChildTypeT:
        size = self.get_size()
        if index >= size:
            raise IndexError(index)
        if len(self._objects) != size:
            self._update_objects()
        return self._objects[index]

    def __setitem__(self, index: int, value):
        child = self[index]
        child.set_state(value)

    def __getattr__(self, name: str):
        alias = self._child_aliases.get(name)
        if alias:
            alias = alias[0]
            alias_obj = self._child_alias_objs.get(name)
            if alias_obj is None:
                obj = self.find_object(alias)
                alias_obj = self._child_alias_objs[name] = _create_child(
                    obj.__class__, None, obj.parent, alias
                )
            return alias_obj
        else:
            return getattr(super(), name)


class Map(SettingsBase[DictStateType]):
    """A ``Map`` object representing key-value settings."""


def _get_new_keywords(obj, *args, **kwds):
    newkwds = {}
    unknown_keywords = set()
    # Convert positional arguments to keyword arguments
    if args:
        argNames = obj.argument_names[:]
        for arg in args:
            argName = argNames.pop(0)
            newkwds[argName] = arg
    if kwds:
        # Convert deprecated keywords through aliases
        # We don't get arguments-aliases from static-info yet.
        argument_aliases_scm = obj.get_attr("arguments-aliases") or {}
        argument_aliases = {}
        for k, v in argument_aliases_scm.items():
            argument_aliases[to_python_name(k)] = to_python_name(v.removeprefix("'"))
        for k, v in kwds.items():
            alias = argument_aliases.get(k)
            if alias:
                newkwds[alias] = v
            elif k in obj.argument_names:
                newkwds[k] = v
            else:
                unknown_keywords.add(k)
    for k in unknown_keywords:
        # Noisily ignore unknown keywords
        warnings.warn(
            f"Unknown keyword '{k}' for command '{obj.python_path}'. "
            "It will be ignored.",
            PyFluentUserWarning,
        )
    return newkwds


class Action(Base):
    """Intermediate Base class for Command and Query class."""

    _child_classes = {}
    _child_aliases = {}
    argument_names = []

    def __init__(self, name: str | None = None, parent=None):
        """__init__ of Action class."""
        super().__init__(name, parent)
        if hasattr(self, "argument_names"):
            for argument in self.argument_names:
                cls = self.__class__._child_classes[argument]
                self._setattr(argument, _create_child(cls, None, self))

    def __dir__(self):
        dir_list = set(list(self.__dict__.keys()) + dir(type(self)))
        return dir_list - set(
            [
                child
                for child in self.argument_names
                if getattr(self, child)._is_deprecated()
            ]
        )

    def get_completer_info(self, prefix="", excluded=None) -> List[List[str]]:
        """Get completer info of all arguments.

        Returns
        -------
        List[List[str]]
            Name, type and docstring of all arguments.
        """
        excluded = excluded or []
        ret = []
        for argument_name in self.argument_names:
            if argument_name not in excluded and argument_name.startswith(prefix):
                argument = getattr(self, argument_name)
                if argument.is_active():
                    ret.append(
                        [
                            argument_name,
                            _get_type_for_completer_info(argument.__class__),
                            argument.__doc__,
                        ]
                    )
        return ret

    def __getattr__(self, name: str):
        alias = self._child_aliases.get(name)
        if alias:
            alias = alias[0]
            alias_obj = self._child_alias_objs.get(name)
            if alias_obj is None:
                obj = self.find_object(alias)
                alias_obj = self._child_alias_objs[name] = _create_child(
                    obj.__class__, None, obj.parent, alias
                )
            return alias_obj
        else:
            return getattr(super(), name)


class BaseCommand(Action):
    """Executes command."""

    def _execute_command(self, *args, **kwds):
        """Execute a command with the specified positional and keyword arguments."""
        if self.flproxy.is_interactive_mode():
            prompt = self.flproxy.get_command_confirmation_prompt(
                self._parent.path, self.obj_name, **kwds
            )
            if prompt:
                valid_responses = {"y": True, "yes": True, "n": False, "no": False}
                while True:
                    response = input(prompt + ": y[es]/n[o] ").strip().lower()
                    if response in valid_responses:
                        if not valid_responses[response]:
                            return
                        break
                    else:
                        print("Please enter 'y[es]' or 'n[o]'.")
        with self._while_executing_command():
            ret = self.flproxy.execute_cmd(self._parent.path, self.obj_name, **kwds)
            if os.getenv("PYFLUENT_NO_FIX_PARAMETER_LIST_RETURN") != "1":
                if (self._parent.path, self.obj_name) in [
                    ("parameters/input-parameters", "list"),
                    ("parameters/output-parameters", "list"),
                ]:
                    ret = _fix_parameter_list_return(ret)
            return ret

    def execute_command(self, *args, **kwds):
        """Execute command."""
        kwds = _get_new_keywords(self, *args, **kwds)
        scmKwds = {}
        for arg, value in kwds.items():
            argument = getattr(self, arg)
            # Convert path-like values for possible file transfer
            value = argument.before_execute(
                command_name=self.python_name, value=value, kwargs=kwds
            )
            # Convert key-value to Scheme key-value
            scmKwds[argument.fluent_name] = argument.to_scheme_keys(
                value,
                argument._root.__class__,
                _get_python_path_comps(argument),
            )
        ret = self._execute_command(*args, **scmKwds)
        for arg, value in kwds.items():
            argument = getattr(self, arg)
            argument.after_execute(
                command_name=self.python_name, value=value, kwargs=kwds
            )
        if (
            self.obj_name in ["create", "make-a-copy"]
            and isinstance(self._parent, NamedObject)
            and ret in self._parent
        ):
            return self._parent[ret]
        return_t = getattr(self, "return_type", None)
        if return_t:
            base_t = _baseTypes.get(return_t)
            if base_t:
                assert_type(ret, base_t._state_type)
            return ret

    def __call__(self, *args, **kwds):
        try:
            return self.execute_command(*args, **kwds)
        except KeyboardInterrupt:
            self._root._on_interrupt(self)
            raise KeyboardInterrupt


# TODO: Remove this after parameter list() method is fixed from Fluent side
def _fix_parameter_list_return(val):
    if isinstance(val, dict):
        new_val = {}
        for name, v in val.items():
            value, units = v
            if len(units) > 0 and isinstance(units[0], str):
                # Symbols are not stripped in the command return in PyConsole.
                # Following code will work in both PyConsole and PyFluent.
                unit = units[0].lstrip("'")
                unit_labels = _fix_parameter_list_return.scheme_eval(
                    f"(units/inquire-available-label-strings-for-quantity '{unit})"
                )
                unit_label = unit_labels[0] if len(unit_labels) > 0 else ""
            else:
                unit_label = ""
            new_val[name] = [value, unit_label]
        return new_val
    return val


_fix_parameter_list_return.scheme_eval = None


class Command(BaseCommand):
    """Command object."""

    def __call__(self, **kwds):
        """Call a command with the specified keyword arguments."""
        try:
            return self.execute_command(**kwds)
        except KeyboardInterrupt:
            self._root._on_interrupt(self)
            raise KeyboardInterrupt


class CommandWithPositionalArgs(BaseCommand):
    """Command Object supporting positional arguments."""

    def __call__(self, *args, **kwds):
        """Call a command with the specified positional and keyword arguments."""
        try:
            return self.execute_command(*args, **kwds)
        except KeyboardInterrupt:
            self._root._on_interrupt(self)
            raise KeyboardInterrupt


class Query(Action):
    """Query object."""

    def __call__(self, **kwds):
        """Call a query with the specified keyword arguments."""
        kwds = _get_new_keywords(self, **kwds)
        scmKwds = {}
        for arg, value in kwds.items():
            argument = getattr(self, arg)
            # Convert key-value to Scheme key-value
            scmKwds[argument.fluent_name] = argument.to_scheme_keys(
                value,
                argument._root.__class__,
                _get_python_path_comps(argument),
            )
        return self.flproxy.execute_query(self._parent.path, self.obj_name, **scmKwds)


_baseTypes = {
    "group": Group,
    "integer": Integer,
    "real": Real,
    "string/symbol": String,
    "string": String,
    "boolean": Boolean,
    "real-list": RealList,
    "integer-list": IntegerList,
    "string-list": StringList,
    "boolean-list": BooleanList,
    "named-object": NamedObject,
    "vector": RealVector,
    "command": Command,
    "query": Query,
    "material-property": String,
    "thread-var": String,
    "list-object": ListObject,
    "file": Filename,
    "file-list": FilenameList,
    "map": Map,
}


def _clean_helpinfo(helpinfo):
    helpinfo = helpinfo.strip("\n").lstrip(" ")
    if not helpinfo.endswith("."):
        helpinfo += "."
    helpinfo = helpinfo[0].upper() + helpinfo[1:]
    return helpinfo


def _fix_help_info(obj_type, helpinfo):
    # The else clause is just picking "object" due to our current
    # knowledge that the implementation only distinguishes between
    # "method" and everything else. This is fragile.
    api_item_type = "method" if obj_type in ("command", "query") else "object"
    fix = _docstrings._fixed_doc_string(api_item_type, helpinfo)
    return fix or helpinfo


class _ChildNamedObjectAccessorMixin(collections.abc.MutableMapping):
    """A mixin class to provide a dictionary interface at a Group class level if the
    Group has multiple named objects of a similar type. For example, boundary conditions
    are grouped by type but quite often we want to access them without the type context.

    The following can be used:
    for name, boundary in setup.boundary_conditions.items():
        print (name, boundary())

    even though actual boundary conditions are stored one level lower to
    boundary_conditions.
    """

    def __getitem__(self, name):
        """Get a child object."""
        for cname in self.child_names:
            cobj = getattr(self, cname)
            try:
                return cobj[name]
            except Exception:
                pass
        raise KeyError(name)

    def __setitem__(self, name, value):
        """Set the state of a child object."""
        self[name].set_state(value)

    def __delitem__(self, name):
        """Delete a child object."""
        for cname in self.child_names:
            cobj = getattr(self, cname)
            try:
                del cobj[name]
                return
            except Exception:
                pass
        raise KeyError(name)

    def __iter__(self):
        """Iterator for child named objects."""
        for cname in self.child_names:
            try:
                for item in getattr(self, cname):
                    yield item
            except Exception:
                continue

    def __len__(self):
        """Number of child named objects."""
        count = 0
        for cname in self.child_names:
            cobj = getattr(self, cname)
            if isinstance(cobj, NamedObject):
                count += len(cobj)
        return count


class CreatableNamedObjectMixin(collections.abc.MutableMapping, Generic[ChildTypeT]):
    """Provides creatable named objects for Fluent 2025 R1 and later."""

    def __setitem__(self, name: str, value):
        if name not in self.get_object_names():
            if self.flproxy.has_wildcard(name):
                child = WildcardPath(
                    self.flproxy,
                    self.path + "/" + name,
                    self.__class__,
                    self.__class__.child_object_type,
                    self,
                )
            else:
                with self._while_creating():
                    self.flproxy.create(self.path, name)
                child = self._create_child_object(name)
        else:
            child = self._objects.get(name)
            if not child:
                child = self._create_child_object(name)
        child.set_state(value)


class CreatableNamedObjectMixinOld(CreatableNamedObjectMixin):
    """Provides creatable named objects for Fluent 2024 R2 and earlier."""

    # In Fluent 2025 R1, the ``create()`` method is available as commands in the ``NamedObject`` class.
    def create(self, name: str = "") -> ChildTypeT:
        """Create a named object.

        Parameters
        ----------
        name: str
            Name of the new object.

        Returns
        -------
        Object
            Object that has been created.
        """
        with self._while_creating():
            self.flproxy.create(self.path, name)
        return self._create_child_object(name)


class _NonCreatableNamedObjectMixin(
    collections.abc.MutableMapping, Generic[ChildTypeT]
):
    def __setitem__(self, name: str, value):
        if name not in self.get_object_names():
            if self.flproxy.has_wildcard(name):
                child = WildcardPath(
                    self.flproxy,
                    self.path + "/" + name,
                    self.__class__,
                    self.__class__.child_object_type,
                    self,
                )
            else:
                raise KeyError(
                    allowed_name_error_message(
                        context=self.python_name,
                        trial_name=name,
                        allowed_values=self.get_object_names(),
                    )
                )
        else:
            child = self._objects.get(name)
            if not child:
                child = self._create_child_object(name)
        child.set_state(value)


class AllowedValuesMixin:
    """Provides allowed values."""

    def allowed_values(self):
        """Get the allowed values of the object."""
        try:
            return self.get_attr(_InlineConstants.allowed_values, (list, str))
        except Exception:
            return []


_bases_by_class = {}


# pylint: disable=missing-raises-doc
def get_cls(name, info, parent=None, version=None, parent_taboo=None):
    """Create a class for the object identified by "path"."""
    try:
        if name == "":
            pname = "root"
        else:
            pname = to_python_name(name)
        obj_type = info["type"]
        base = _baseTypes.get(obj_type)
        if obj_type == "command" and name in ["create", "rename", "delete", "resize"]:
            base = CommandWithPositionalArgs
        if base is None:
            settings_logger.warning(
                f"Unable to find base class for '{name}' "
                f"(type = '{obj_type}'). "
                f"Falling back to String."
            )
            base = String
        dct = {"fluent_name": name, "_version": version}
        helpinfo = info.get("help")
        if helpinfo:
            dct["__doc__"] = _fix_help_info(obj_type, _clean_helpinfo(helpinfo))
        else:
            if parent is None:
                dct["__doc__"] = "'root' object."
            else:
                if obj_type == "command":
                    dct["__doc__"] = f"'{pname.strip('_')}' command."
                elif obj_type == "query":
                    dct["__doc__"] = f"'{pname.strip('_')}' query."
                else:
                    dct["__doc__"] = f"'{pname.strip('_')}' child."

        include_child_named_objects = info.get(
            "include-child-named-objects?", False
        ) or info.get("include_child_named_objects", False)
        user_creatable = info.get("user-creatable?", False) or info.get(
            "user_creatable", False
        )

        if version == "222":
            user_creatable = True

        bases = (base,)
        if include_child_named_objects:
            bases = bases + (_ChildNamedObjectAccessorMixin,)
        if obj_type == "named-object" and user_creatable:
            if version < "251":
                bases = bases + (CreatableNamedObjectMixinOld,)
            else:
                bases = bases + (CreatableNamedObjectMixin,)
        elif obj_type == "named-object":
            bases = bases + (_NonCreatableNamedObjectMixin,)
        elif info.get("has-allowed-values"):
            bases += (AllowedValuesMixin,)
        elif info.get("file_purpose") == "input":
            bases += (_InputFile,)
        elif info.get("file_purpose") == "output":
            bases += (_OutputFile,)
        elif info.get("file_purpose") == "inout":
            bases += (_InOutFile,)

        original_pname = pname
        i = 1
        if parent_taboo:
            while pname in parent_taboo:
                pname = f"{original_pname}_{i}"
                i += 1
        parent_attr_name = pname
        if info.get("file_purpose"):  # not generalizing for performance
            while pname in _bases_by_class and _bases_by_class[pname] != bases:
                pname = f"{original_pname}_{i}"
                i += 1
            _bases_by_class[pname] = bases
        if parent_taboo:
            parent_taboo.add(pname)

        dct["_child_classes"] = {}
        cls = type(pname, bases, dct)

        taboo = set(dir(cls))
        taboo |= set(
            [
                "child_names",
                "command_names",
                "query_names",
                "argument_names",
                "child_object_type",
            ]
        )

        doc = ""

        def _process_cls_names(info_dict, names, write_doc=False):
            nonlocal taboo
            nonlocal cls

            for cname, cinfo in info_dict.items():
                ccls, parent_attr_name = get_cls(
                    cname, cinfo, cls, version=version, parent_taboo=taboo
                )

                if write_doc:
                    nonlocal doc
                    th = ccls._state_type
                    th = th.__name__ if hasattr(th, "__name__") else str(th)
                    doc += f"    {ccls.__name__} : {th}\n"
                    doc += f"        {ccls.__doc__}\n"

                names.append(parent_attr_name)
                taboo.add(ccls.__name__)
                cls._child_classes[parent_attr_name] = ccls

        children = info.get("children")
        if children:
            taboo.add("child_names")
            cls.child_names = []
            _process_cls_names(children, cls.child_names)

        commands = info.get("commands")
        if commands:
            commands.pop("exit", None)
            commands.pop("switch-to-meshing-mode", None)
        if commands and not user_creatable:
            commands.pop("create", None)
        if commands:
            cls.command_names = []
            _process_cls_names(commands, cls.command_names)

        queries = info.get("queries")
        if queries:
            cls.query_names = []
            _process_cls_names(queries, cls.query_names)

        arguments = info.get("arguments")
        if arguments:
            doc = cls.__doc__
            doc += "\n\n"
            doc += "Parameters\n"
            doc += "----------\n"
            cls.argument_names = []
            _process_cls_names(arguments, cls.argument_names, write_doc=True)
            cls.__doc__ = doc

        if version < "242":
            cls.return_type = "object"
        else:
            return_type = info.get("return-type") or info.get("return_type")
            if return_type:
                cls.return_type = return_type

        object_type = info.get("object-type", False) or info.get("object_type", False)
        if object_type:
            cls.child_object_type, _ = get_cls(
                "child-object-type", object_type, cls, version=version
            )
            cls.child_object_type.get_name = lambda self: self._name

        child_aliases = info.get("child-aliases") or info.get("child_aliases", {})
        command_aliases = info.get("command-aliases") or info.get("command_aliases", {})
        query_aliases = info.get("query-aliases") or info.get("query_aliases", {})
        argument_aliases = info.get("arguments-aliases") or info.get(
            "arguments_aliases", {}
        )
        if child_aliases or command_aliases or query_aliases or argument_aliases:
            cls._child_aliases = {}
            # No need to differentiate in the Python implementation
            for k, v in (
                child_aliases | command_aliases | query_aliases | argument_aliases
            ).items():
                # Storing the original name as we don't have any other way
                # to recover it at runtime.
                cls._child_aliases[to_python_name(k)] = (
                    "/".join(
                        x if x == ".." else to_python_name(x) for x in v.split("/")
                    ),
                    k,
                )

    except Exception:
        print(
            f"Unable to construct class for '{name}' of "
            f"'{parent.fluent_name if parent else None}'"
        )
        raise
    return cls, parent_attr_name


def _gethash(obj_info):
    dhash = hashlib.sha256()
    dhash.update(pickle.dumps(obj_info))
    return dhash.hexdigest()


def get_root(
    flproxy,
    version: str = "",
    interrupt: Any | None = None,
    file_transfer_service: Any | None = None,
    scheme_eval=None,
) -> Group:
    """Get the root settings object.

    Parameters
    ----------
    flproxy: Proxy
        Object that interfaces with the Fluent backend.
    interrupt: optional
        To interrupt interruptible commands.
    file_transfer_service : optional
        File transfer service. Uploads/downloads files to/from the server.
    scheme_eval : Any
        A gRPC service to execute Scheme code.
    version : str
        Fluent version.

    Returns
    -------
    root object

    Raises
    ------
    RuntimeError
        If hash values are inconsistent.
    """
    from ansys.fluent.core import CODEGEN_OUTDIR, utils

    if os.getenv("PYFLUENT_USE_RUNTIME_PYTHON_CLASSES") == "1":
        obj_info = flproxy.get_static_info()
        root_cls, _ = get_cls("", obj_info, version=version)
    else:
        try:
            settings = utils.load_module(
                f"settings_{version}",
                CODEGEN_OUTDIR / "solver" / f"settings_{version}.py",
            )
            root_cls = settings.root
        except FileNotFoundError:
            obj_info = flproxy.get_static_info()
            root_cls, _ = get_cls("", obj_info, version=version)
    root = root_cls()
    root.set_flproxy(flproxy)
    root._set_on_interrupt(interrupt)
    root._set_file_transfer_service(file_transfer_service)
    _Alias.scheme_eval = scheme_eval
    _fix_parameter_list_return.scheme_eval = scheme_eval
    root._setattr("_file_transfer_service", file_transfer_service)
    return root


def find_children(obj, identifier="*"):
    """Returns path of all the child objects matching an identifier.

    Parameters
    ----------
    obj: Object
        Object whose children need to be queried.
    identifier: str
        Identifier to find specific children.

    Returns
    -------
    List
    """
    list_of_children = []
    _list_children(obj.__class__, identifier, [], list_of_children)
    return list_of_children


def _list_children(cls, identifier, path, list_of_children):
    if issubclass(cls, (NamedObject, ListObject)):
        if hasattr(cls.child_object_type, "child_names"):
            _get_child_path(cls.child_object_type, path, identifier, list_of_children)
    if issubclass(cls, Group):
        _get_child_path(cls, path, identifier, list_of_children)


def _get_child_path(cls, path, identifier, list_of_children):
    for name in cls.child_names:
        path.append(name)
        if fnmatch.fnmatch(name, identifier):
            path_to_append = "/".join(path)
            if path_to_append not in list_of_children:
                list_of_children.append(path_to_append)
        _list_children(cls._child_classes[name], identifier, path, list_of_children)
        path.pop()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""

Modeling Cavitation
-------------------
This example examines the pressure-driven cavitating flow of water through a
sharp-edged orifice. This is a typical configuration in fuel injectors, and
brings a challenge to the physics and numerics of cavitation models because of
the high pressure differentials involved and the high ratio of liquid to vapor
density.

This example uses the multiphase modeling capability of Ansys Fluent, and will
be able to predict the strong cavitation near the orifice after flow separation
at a sharp edge.

**Workflow tasks**

This example demonstrates how to do the following:

- Set boundary conditions for internal flow
- Use the mixture model with cavitation effects
- Calculate a solution using the pressure-based coupled solver

**Problem description**

The problem considers the cavitation caused by the flow separation after a
sharp-edged orifice. The flow is pressure driven, with an inlet pressure of 500
kPa and an outlet pressure of 95 kPa. The orifice diameter is 4 mm, and the
geometrical parameters of the orifice are D/d = 2.88 and L/d = 4, where D, d,
and L are the inlet diameter, orifice diameter, and orifice length respectively.
"""

###############################################################################
# Example Setup
# -------------
# Before you can begin, you must set up the example and initialize this
# workflow.
#
# Launch Fluent
# ~~~~~~~~~~~~~
# Perform required imports, which includes downloading and importing
# the geometry file.

# sphinx_gallery_thumbnail_path = '_static/cavitation_model_thumb.png'

import ansys.fluent.core as pyfluent
from ansys.fluent.core import examples

cav_file = examples.download_file("cav.msh.gz", "pyfluent/cavitation")

###############################################################################
# Launch a Fluent session in the 2d solution mode with double precision running
# on four processors and print Fluent version.

solver = pyfluent.launch_fluent(
    precision="double",
    processor_count=4,
    mode="solver",
    version="2d",
)
print(solver.get_fluent_version())

###############################################################################
# Read the mesh that was downloaded.

solver.file.read_mesh(file_name=cav_file)

solver.mesh.check()

###############################################################################
# Specify an axisymmetric model.

solver.setup.general.solver.two_dim_space = "axisymmetric"

###############################################################################
# Enable the multiphase mixture model.

solver.setup.models.multiphase.models = "mixture"

solver.tui.define.models.multiphase.mixture_parameters("no", "implicit")

###############################################################################
# Enable the k-ω SST turbulence model.

solver.setup.models.viscous = {
    "model": "k-omega",
    "k_omega_model": "sst",
}

###############################################################################
# Define materials
# ~~~~~~~~~~~~~~~~
# Create a material named water, using 1000 kg/m3 for density and 0.001 kg/m–s
# for viscosity. Then, copy water vapor properties from the database and modify
# the copy by changing the density to 0.02558 kg/m3 and the viscosity to
# 1.26e-06 kg/m–s.

solver.setup.materials.fluid["water"] = {
    "density": {
        "option": "constant",
        "value": 1000,
    },
    "viscosity": {
        "option": "constant",
        "value": 0.001,
    },
}

solver.setup.materials.database.copy_by_name(type="fluid", name="water-vapor")

solver.setup.materials.fluid["water-vapor"] = {
    "density": {"value": 0.02558},
    "viscosity": {"value": 1.26e-06},
}

###############################################################################
# Phases
# ~~~~~~
# Change the name of the primary phase to "liquid" and the secondary phase to
# "water-vapor". Then, enable the cavitation model and set the number of mass
# transfer mechanisms to 1. Finally, specify cavitation as a mass transfer
# mechanism occurring from the liquid to the vapor.

solver.tui.define.phases.set_domain_properties.change_phases_names("vapor", "liquid")

solver.tui.define.phases.set_domain_properties.phase_domains.liquid.material(
    "yes", "water"
)

solver.tui.define.phases.set_domain_properties.phase_domains.vapor.material(
    "yes", "water-vapor"
)

solver.tui.define.phases.set_domain_properties.interaction_domain.heat_mass_reactions.mass_transfer(
    1, "liquid", "vapor", "cavitation", "1", "no", "no", "no"
)

###############################################################################
# Boundary Conditions
# ~~~~~~~~~~~~~~~~~~~
# For the first inlet momentum boundary conditions set the direction
# specification method to 'normal to boundary', gauge total pressure as 500 kPa
# and supersonic or initial gauge pressure as 449 kPa.
#
# For the turbulence settings choose 'Intensity and Viscosity Ratio' for
# turbulent specification. Set turbulent intensity and turbulent viscosity
# ratio to 0.05 and 10 respectively.

inlet_1 = solver.setup.boundary_conditions.pressure_inlet["inlet_1"].phase

inlet_1["mixture"] = {
    "momentum": {
        "gauge_total_pressure": {"value": 500000},
        "supersonic_or_initial_gauge_pressure": {"value": 449000},
        "direction_specification_method": "Normal to Boundary",
    },
    "turbulence": {
        "turbulent_specification": "Intensity and Viscosity Ratio",
        "turbulent_intensity": 0.05,
        "turbulent_viscosity_ratio": 10,
    },
}

###############################################################################
# Before copying inlet_1's boundary conditions to inlet_2, set the vapor fraction
# to 0.

inlet_1["vapor"] = {
    "multiphase": {
        "volume_fraction": {"value": 0},
    },
}

solver.setup.boundary_conditions.copy(from_="inlet_1", to="inlet_2")

###############################################################################
# For the outlet boundary conditions, set the gauge pressure as 95 kPa. Use
# the same turbulence and volume fraction settings as the inlets.

outlet = solver.setup.boundary_conditions.pressure_outlet["outlet"].phase

outlet["mixture"] = {
    "momentum": {
        "gauge_pressure": {"value": 95000},
    },
    "turbulence": {
        "turbulent_specification": "Intensity and Viscosity Ratio",
        "turbulent_intensity": 0.04,
        "turbulent_viscosity_ratio": 10,
    },
}

outlet["vapor"] = {
    "multiphase": {
        "volume_fraction": {"value": 0},
    },
}

###############################################################################
# Operating Conditions
# ~~~~~~~~~~~~~~~~~~~~
# Set the operating pressure to 0.

solver.setup.general.operating_conditions.operating_pressure = 0

###############################################################################
# Solution
# ~~~~~~~~
# To configure the discretization scheme, set 'first order
# upwind' method for turbulent kinetic energy and turbulent dissipation rate, 'quick'
# for the momentum and volume fraction, and 'presto!' for pressure.

methods = solver.solution.methods

methods.discretization_scheme = {
    "k": "first-order-upwind",
    "mom": "quick",
    "mp": "quick",
    "omega": "first-order-upwind",
    "pressure": "presto!",
}

###############################################################################
# For the pressure velocity coupling scheme choose 'Coupled'. Set the pseudo
# time step method to 'global time step' and enable 'High Order Term Relaxation'.
# Then, set the explicit relaxation factor for 'Volume Fraction' to
# 0.3.

methods.p_v_coupling.flow_scheme = "Coupled"

methods.pseudo_time_method.formulation.coupled_solver = "global-time-step"

methods.high_order_term_relaxation.enable = True

solver.solution.controls.pseudo_time_explicit_relaxation_factor.global_dt_pseudo_relax[
    "mp"
] = 0.3

###############################################################################
# To plot the residuals, enable plotting and set the convergence criteria to
# 1e-05 for x-velocity, y-velocity, k, omega, and vf-vapor. Enable the specified
# initial pressure then initialize the solution with hybrid initialization.

resid_eqns = solver.solution.monitor.residual.equations

resid_eqns["continuity"].absolute_criteria = 1e-5
resid_eqns["x-velocity"].absolute_criteria = 1e-5
resid_eqns["y-velocity"].absolute_criteria = 1e-5
resid_eqns["k"].absolute_criteria = 1e-5
resid_eqns["omega"].absolute_criteria = 1e-5
resid_eqns["vf-vapor"].absolute_criteria = 1e-5

initialization = solver.solution.initialization

initialization.initialization_type = "hybrid"
initialization.hybrid_init_options.general_settings.initial_pressure = True
initialization.hybrid_initialize()

###############################################################################
# Save and Run
# ~~~~~~~~~~~~
# Save the case file 'cav.cas.h5'. Then, start the calculation by requesting
# 500 iterations. Save the final case file and the data.

solver.file.write_case(file_name="cav.cas.h5")

solver.solution.run_calculation.iterate(iter_count=500)

solver.file.write_case_data(file_name="cav.cas.h5")

###############################################################################
# Post Processing
# ~~~~~~~~~~~~~~~
# Since Fluent is being run without the GUI, we will need to export plots as
# picture files. Edit the picture settings to use a custom resolution so that
# the images are large enough.

graphics = solver.results.graphics
# use_window_resolution option not active inside containers or Ansys Lab environment
if graphics.picture.use_window_resolution.is_active():
    graphics.picture.use_window_resolution = False
graphics.picture.x_resolution = 1920
graphics.picture.y_resolution = 1440

###############################################################################
# Create contour plots
# ~~~~~~~~~~~~~~~~~~~~
# Create a contour plot for static pressure, turbulent kinetic energy and the
# volume fraction of water vapor. For each plot enable banded coloring and
# filled option.

graphics = solver.results.graphics

graphics.contour["contour_static_pressure"] = {
    "coloring": {
        "option": "banded",
        "smooth": False,
    },
    "field": "pressure",
    "filled": True,
}

###############################################################################
# Mirror the display around the symmetry plane to show the full model.

solver.settings.results.graphics.views.mirror_zones = ["symm_2", "symm_1"]

graphics.contour["contour_static_pressure"].display()

graphics.picture.save_picture(file_name="contour_static_pressure.png")

###############################################################################
# .. image:: /_static/cavitation_model_012.png
#   :width: 500pt
#   :align: center

graphics.contour.create("contour_tke")

graphics.contour["contour_tke"] = {
    "coloring": {
        "option": "banded",
        "smooth": False,
    },
    "field": "turb-kinetic-energy",
    "filled": True,
}

graphics.contour["contour_tke"].display()

graphics.picture.save_picture(file_name="contour_tke.png")

###############################################################################
# .. image:: /_static/cavitation_model_011.png
#   :width: 500pt
#   :align: center

graphics.contour.create("contour_vf_vapor")

graphics.contour["contour_vf_vapor"] = {
    "coloring": {
        "option": "banded",
        "smooth": False,
    },
    "field": "vapor-vof",
    "filled": True,
}

graphics.contour["contour_vf_vapor"].display()

graphics.picture.save_picture(file_name="contour_vf_vapor.png")

###############################################################################
# .. image:: /_static/cavitation_model.png
#   :width: 500pt
#   :align: center

# Save case to 'cav.cas.h5' and exit

solver.settings.file.write_case(file_name="cav.cas.h5")

solver.exit()



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from functools import partial

import pytest
from util.meshing_workflow import (
    assign_task_arguments,
    execute_task_with_pre_and_postcondition_checks,
)

from ansys.fluent.core import examples
from ansys.fluent.core.utils.fluent_version import FluentVersion


@pytest.mark.fluent_version(">=23.1")
@pytest.mark.nightly
@pytest.mark.codegen_required
def test_mixing_elbow_meshing_workflow(
    watertight_workflow_session,
    mixing_elbow_geometry_filename,
):
    """This test covers generic meshing workflow behaviour."""
    meshing_session = watertight_workflow_session
    workflow = meshing_session.workflow

    ###############################################################################

    assign_task_args = partial(
        assign_task_arguments, workflow=workflow, check_state=True
    )

    execute_task_with_pre_and_postconditions = partial(
        execute_task_with_pre_and_postcondition_checks,
        workflow=workflow,
    )

    ###############################################################################
    # Import the CAD geometry
    # Query the task state before and after task execution
    assign_task_args(
        task_name="Import Geometry",
        FileName=mixing_elbow_geometry_filename,
        LengthUnit="in",
    )

    execute_task_with_pre_and_postconditions(task_name="Import Geometry")

    ###############################################################################
    # Add local sizing
    # Query the task state before and after task execution
    workflow.TaskObject["Add Local Sizing"].AddChildToTask()

    execute_task_with_pre_and_postconditions(task_name="Add Local Sizing")

    ###############################################################################
    # Generate the surface mesh
    # Query the task state before and after task execution
    assign_task_args(
        task_name="Generate the Surface Mesh", CFDSurfaceMeshControls={"MaxSize": 0.3}
    )

    execute_task_with_pre_and_postconditions(task_name="Generate the Surface Mesh")

    ###############################################################################
    # Describe the geometry
    # Query the task state before and after task execution
    describe_geo = workflow.TaskObject["Describe Geometry"]
    describe_geo.UpdateChildTasks(SetupTypeChanged=False)
    assign_task_args(
        task_name="Describe Geometry",
        SetupType="The geometry consists of only fluid regions with no voids",
    )
    describe_geo.UpdateChildTasks(SetupTypeChanged=True)

    execute_task_with_pre_and_postconditions(task_name="Describe Geometry")

    ###############################################################################
    # Update Boundaries Task
    # Query the task state before and after task execution
    assign_task_args(
        task_name="Update Boundaries",
        BoundaryLabelList=["wall-inlet"],
        BoundaryLabelTypeList=["wall"],
        OldBoundaryLabelList=["wall-inlet"],
        OldBoundaryLabelTypeList=["velocity-inlet"],
    )

    execute_task_with_pre_and_postconditions(task_name="Update Boundaries")

    ###############################################################################
    # Update your regions
    # Query the task state before and after task execution

    execute_task_with_pre_and_postconditions(task_name="Update Regions")

    ###############################################################################
    # Add Boundary Layers
    # Query the task state before and after task execution
    add_boundary_layer = workflow.TaskObject["Add Boundary Layers"]
    add_boundary_layer.AddChildToTask()
    add_boundary_layer.InsertCompoundChildTask()
    assign_task_args(
        task_name="smooth-transition_1", BLControlName="smooth-transition_1"
    )
    add_boundary_layer.Arguments = {}

    execute_task_with_pre_and_postconditions(task_name="Add Boundary Layers")

    ###############################################################################
    # Generate the volume mesh
    # Query the task state before and after task execution
    assign_task_args(
        task_name="Generate the Volume Mesh",
        VolumeFill="poly-hexcore",
        VolumeFillControls={"HexMaxCellLength": 0.3},
    )

    execute_task_with_pre_and_postconditions(task_name="Generate the Volume Mesh")

    ###############################################################################
    # Check the mesh in Meshing mode
    # TODO: Remove the if condition after a stable version of 23.1 is available and update the commands as required.
    if meshing_session.get_fluent_version() < FluentVersion.v231:
        meshing_session.tui.mesh.check_mesh()


@pytest.mark.codegen_required
def test_meshing_workflow_raises_exception_on_invalid_task_name(
    watertight_workflow_session,
):
    try:
        watertight_workflow_session.workflow.TaskObject["no such task"]
    except Exception:
        pass
    else:
        assert False


"""
Cannot enable this test because meshing workflow makes invalid queries as
soon as the meshing application is started:

Error: workflow/cx-create-workflow-tree:Invalid query for child TaskType from parent /Workflow
Error Object: ()


def test_meshing_workflow_raises_exception_on_invalid_key_in_task_args(
    model_object_throws_on_invalid_arg,
    watertight_workflow_session,
    mixing_elbow_geometry_filename
):
    # task_names = ("Import Geometry", "Add Local Sizing")
    task_names = ("Add Local Sizing",)
    for task_name in task_names:
        task = watertight_workflow_session.workflow.TaskObject[task_name]
        try:
            task.Arguments = {"no such arg": 42}
        except Exception:
            pass
        else:
            assert False

def test_meshing_workflow_raises_exception_on_invalid_key_in_task_args_2(
    model_object_throws_on_invalid_arg,
    watertight_workflow_session,
    mixing_elbow_geometry_filename
):
    workflow = watertight_workflow_session.workflow
    assign_task_args = partial(
        assign_task_arguments, workflow=workflow, check_state=False
    )

    assign_task_args(
        task_name="Import Geometry", FileName=mixing_elbow_geometry_filename, LengthUnit="in"
    )

    workflow.TaskObject["Import Geometry"].Execute()

    try:
        assign_task_args(
            task_name="Add Local Sizing", XXX=42
        )
    except:
        pass
    else:
        assert False
"""


@pytest.mark.skip("Wait for later implementation.")
@pytest.mark.fluent_version(">=23.1")
@pytest.mark.codegen_required
def test_read_only_behaviour_of_command_arguments(new_meshing_session):
    session_new = new_meshing_session
    w = session_new.workflow
    m = session_new.meshing.ImportGeometry.create_instance
    w.InitializeWorkflow(WorkflowType="Watertight Geometry")
    import_geom = w.TaskObject["Import Geometry"]

    with pytest.raises(AttributeError):
        import_geom.arguments.MeshUnit.set_state("in")

    with pytest.raises(AttributeError):
        import_geom.arguments.CadImportOptions.OneZonePer.set_state(None)

    assert "set_state" in dir(m())
    assert "set_state" in dir(m().NumParts)


@pytest.mark.codegen_required
def test_dummy_journal_data_model_methods(new_meshing_session):
    session_new = new_meshing_session
    w = session_new.workflow
    w.InitializeWorkflow(WorkflowType="Watertight Geometry")
    import_geom = w.TaskObject["Import Geometry"]

    with pytest.raises(AttributeError):
        import_geom.delete_child()


@pytest.mark.codegen_required
def test_iterate_meshing_workflow_task_container(new_meshing_session):
    workflow = new_meshing_session.workflow
    workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    tasks = [task for task in workflow.TaskObject]
    assert len(tasks) == 11
    assert tasks[0].name() == "Import Geometry"


@pytest.mark.codegen_required
def test_nonexistent_attrs(new_meshing_session):
    meshing = new_meshing_session
    assert not hasattr(meshing.workflow, "xyz")
    with pytest.raises(AttributeError):
        meshing.workflow.xyz


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=23.2")
def test_old_workflow_structure(new_meshing_session):
    meshing = new_meshing_session
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    assert meshing.workflow.TaskObject["Import Geometry"]
    with pytest.raises(AttributeError):
        meshing.workflow.import_geometry


@pytest.mark.nightly
@pytest.mark.codegen_required
@pytest.mark.fluent_version("==24.2")
def test_new_2d_meshing_workflow(new_meshing_session):
    # Import geometry
    import_file_name = examples.download_file("NACA0012.fmd", "pyfluent/airfoils")
    meshing = new_meshing_session
    meshing.workflow.InitializeWorkflow(WorkflowType="2D Meshing")
    meshing.workflow.TaskObject["Load CAD Geometry"].Arguments.set_state(
        {
            r"FileName": import_file_name,
            r"LengthUnit": r"mm",
            r"Refaceting": {
                r"Refacet": False,
            },
        }
    )
    meshing.workflow.TaskObject["Load CAD Geometry"].Execute()

    meshing.workflow.TaskObject["Update Regions"].Execute()
    meshing.workflow.TaskObject["Update Boundaries"].Arguments.set_state(
        {
            r"SelectionType": r"zone",
        }
    )
    meshing.workflow.TaskObject["Update Boundaries"].Execute()

    meshing.workflow.TaskObject["Define Global Sizing"].Arguments.set_state(
        {
            r"CurvatureNormalAngle": 20,
            r"MaxSize": 2000,
            r"MinSize": 5,
            r"SizeFunctions": r"Curvature",
        }
    )
    meshing.workflow.TaskObject["Define Global Sizing"].Execute()

    meshing.workflow.TaskObject["Add Local Sizing"].Arguments.set_state(
        {
            r"AddChild": r"yes",
            r"BOIControlName": r"boi_1",
            r"BOIExecution": r"Body Of Influence",
            r"BOIFaceLabelList": [r"boi"],
            r"BOISize": 50,
            r"BOIZoneorLabel": r"label",
            r"DrawSizeControl": True,
        }
    )
    meshing.workflow.TaskObject["Add Local Sizing"].AddChildAndUpdate(DeferUpdate=False)

    meshing.workflow.TaskObject["Add Local Sizing"].Arguments.set_state(
        {
            r"AddChild": r"yes",
            r"BOIControlName": r"edgesize_1",
            r"BOIExecution": r"Edge Size",
            r"BOISize": 5,
            r"BOIZoneorLabel": r"label",
            r"DrawSizeControl": True,
            r"EdgeLabelList": [r"airfoil-te"],
        }
    )
    meshing.workflow.TaskObject["Add Local Sizing"].AddChildAndUpdate(DeferUpdate=False)

    meshing.workflow.TaskObject["Add Local Sizing"].Arguments.set_state(
        {
            r"AddChild": r"yes",
            r"BOIControlName": r"curvature_1",
            r"BOICurvatureNormalAngle": 10,
            r"BOIExecution": r"Curvature",
            r"BOIMaxSize": 2,
            r"BOIMinSize": 1.5,
            r"BOIScopeTo": r"edges",
            r"BOIZoneorLabel": r"label",
            r"DrawSizeControl": True,
            r"EdgeLabelList": [r"airfoil"],
        }
    )
    meshing.workflow.TaskObject["Add Local Sizing"].AddChildAndUpdate(DeferUpdate=False)

    meshing.workflow.TaskObject["Add 2D Boundary Layers"].Arguments.set_state(
        {
            r"AddChild": r"yes",
            r"BLControlName": r"aspect-ratio_1",
            r"NumberOfLayers": 4,
            r"OffsetMethodType": r"aspect-ratio",
        }
    )
    meshing.workflow.TaskObject["Add 2D Boundary Layers"].AddChildAndUpdate(
        DeferUpdate=False
    )
    meshing.workflow.TaskObject["Generate the Surface Mesh"].Arguments.set_state(
        {
            r"Surface2DPreferences": {
                r"MergeEdgeZonesBasedOnLabels": r"no",
                r"MergeFaceZonesBasedOnLabels": r"no",
                r"ShowAdvancedOptions": True,
            },
        }
    )
    meshing.workflow.TaskObject["Generate the Surface Mesh"].Execute()

    meshing.workflow.TaskObject["aspect-ratio_1"].Revert()
    meshing.workflow.TaskObject["aspect-ratio_1"].Arguments.set_state(
        {
            r"AddChild": r"yes",
            r"BLControlName": r"uniform_1",
            r"FirstLayerHeight": 2,
            r"NumberOfLayers": 4,
            r"OffsetMethodType": r"uniform",
        }
    )
    meshing.workflow.TaskObject["aspect-ratio_1"].Execute()
    meshing.workflow.TaskObject["Generate the Surface Mesh"].Arguments.set_state(None)
    meshing.workflow.TaskObject["Generate the Surface Mesh"].Arguments.set_state(
        {
            r"Surface2DPreferences": {
                r"MergeEdgeZonesBasedOnLabels": r"no",
                r"MergeFaceZonesBasedOnLabels": r"no",
                r"ShowAdvancedOptions": True,
            },
        }
    )
    meshing.workflow.TaskObject["Generate the Surface Mesh"].Execute()
    meshing.workflow.TaskObject["uniform_1"].Revert()
    meshing.workflow.TaskObject["uniform_1"].Arguments.set_state(
        {
            r"AddChild": r"yes",
            r"BLControlName": r"smooth-transition_1",
            r"FirstLayerHeight": 2,
            r"NumberOfLayers": 7,
            r"OffsetMethodType": r"smooth-transition",
        }
    )
    meshing.workflow.TaskObject["uniform_1"].Execute()

    meshing.workflow.TaskObject["Generate the Surface Mesh"].Arguments.set_state(None)
    meshing.workflow.TaskObject["Generate the Surface Mesh"].Arguments.set_state(
        {
            r"Surface2DPreferences": {
                r"MergeEdgeZonesBasedOnLabels": r"no",
                r"MergeFaceZonesBasedOnLabels": r"no",
                r"ShowAdvancedOptions": True,
            },
        }
    )
    meshing.workflow.TaskObject["Generate the Surface Mesh"].Execute()

    meshing.workflow.TaskObject["Export Fluent 2D Mesh"].Arguments.set_state(
        {
            r"FileName": r"C:\ANSYSDev\PyFluent_Dev_01\pyfluent\out\case1.msh.h5",
        }
    )
    meshing.workflow.TaskObject["Export Fluent 2D Mesh"].Execute()

    # Switch to solution mode
    solver = meshing.switch_to_solver()
    assert solver


@pytest.mark.fluent_version(">=24.1")
def test_setting_none_type_tasks(new_meshing_session):
    meshing = new_meshing_session
    meshing.workflow.InitializeWorkflow(WorkflowType=r"Fault-tolerant Meshing")
    meshing.workflow.setState(
        {
            r"TaskObject:Describe Overset Features": {},
        }
    )
    meshing.workflow.TaskObject["Describe Overset Features"].CommandName.setState(
        r"OversetOptions"
    )

    assert (
        meshing.workflow.TaskObject["Describe Overset Features"].CommandName()
        == "DescribeOversetFeatures"
    )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing class encapsulating Fluent connection."""

from asyncio import Future
import functools
import logging
import threading
from typing import Any, Dict
import warnings
import weakref

from ansys.api.fluent.v0 import svar_pb2 as SvarProtoModule
import ansys.fluent.core as pyfluent
from ansys.fluent.core.pyfluent_warnings import PyFluentDeprecationWarning
from ansys.fluent.core.services import SchemeEval, service_creator
from ansys.fluent.core.services.field_data import ZoneInfo, ZoneType
from ansys.fluent.core.services.reduction import ReductionService
from ansys.fluent.core.services.solution_variables import (
    SolutionVariableData,
    SolutionVariableInfo,
)
from ansys.fluent.core.session import BaseSession
from ansys.fluent.core.session_shared import _make_datamodel_module, _make_tui_module
from ansys.fluent.core.solver import flobject
from ansys.fluent.core.solver.flobject import (
    DeprecatedSettingWarning,
    Group,
    NamedObject,
    SettingsBase,
    StateT,
    StateType,
)
import ansys.fluent.core.solver.function.reduction as reduction_old
from ansys.fluent.core.streaming_services.events_streaming import SolverEvent
from ansys.fluent.core.streaming_services.monitor_streaming import MonitorsManager
from ansys.fluent.core.system_coupling import SystemCoupling
from ansys.fluent.core.utils.execution import asynchronous
from ansys.fluent.core.utils.fluent_version import (
    FluentVersion,
    get_version_for_file_name,
)
from ansys.fluent.core.workflow import ClassicWorkflow

tui_logger = logging.getLogger("pyfluent.tui")
datamodel_logger = logging.getLogger("pyfluent.datamodel")


def _set_state_safe(obj: SettingsBase, state: StateType):
    try:
        obj.set_state(state)
    except RuntimeError:
        if isinstance(obj, NamedObject):
            for k, v in state.items():
                _set_state_safe(obj[k], v)
        elif isinstance(obj, Group):
            for k, v in state.items():
                _set_state_safe(getattr(obj, k), v)
        else:
            datamodel_logger.debug(f"set_state failed at {obj.path}")


def _import_settings_root(root):
    _class_dict = {}
    api_keys = []
    if hasattr(root, "child_names"):
        api_keys = root.child_names

    for root_item in api_keys:
        _class_dict[root_item] = root.__dict__[root_item]

    settings_api_root = type("SettingsRoot", (object,), _class_dict)
    return settings_api_root()


class Solver(BaseSession):
    """Encapsulates a Fluent solver session.

    A ``tui`` object for solver TUI
    commanding, and solver settings objects are all exposed here.
    """

    def __init__(
        self,
        fluent_connection,
        scheme_eval: SchemeEval,
        file_transfer_service: Any | None = None,
        start_transcript: bool = True,
        launcher_args: Dict[str, Any] | None = None,
    ):
        """Solver session.

        Parameters
        ----------
        fluent_connection (:ref:`ref_fluent_connection`):
            Encapsulates a Fluent connection.
        scheme_eval: SchemeEval
            Instance of ``SchemeEval`` to execute Fluent's scheme code on.
        file_transfer_service : Optional
            Service for uploading and downloading files.
        start_transcript : bool, optional
            Whether to start the Fluent transcript in the client.
            The default is ``True``, in which case the Fluent
            transcript can be subsequently started and stopped
            using method calls on the ``Session`` object.
        """
        super(Solver, self).__init__(
            fluent_connection=fluent_connection,
            scheme_eval=scheme_eval,
            file_transfer_service=file_transfer_service,
            start_transcript=start_transcript,
            launcher_args=launcher_args,
            event_type=SolverEvent,
            get_zones_info=weakref.WeakMethod(self._get_zones_info),
        )
        self._build_from_fluent_connection(fluent_connection, scheme_eval)

    def _build_from_fluent_connection(
        self,
        fluent_connection,
        scheme_eval: SchemeEval,
        file_transfer_service: Any | None = None,
    ):
        self._tui_service = self._datamodel_service_tui
        self._se_service = self._datamodel_service_se
        self._tui = None
        self._workflow = None
        self._system_coupling = None
        self._settings_root = None
        self._fluent_version = None
        self._lck = threading.Lock()
        self._solution_variable_service = service_creator("svar").create(
            fluent_connection._channel, fluent_connection._metadata
        )
        self.fields.solution_variable_info = SolutionVariableInfo(
            self._solution_variable_service
        )
        self._reduction_service = self._fluent_connection.create_grpc_service(
            ReductionService, self._error_state
        )
        if FluentVersion(self._version) >= FluentVersion.v241:
            self.fields.reduction = service_creator("reduction").create(
                self._reduction_service, self
            )
        else:
            self.fields.reduction = reduction_old
        self._settings_api_root = None
        self.fields.solution_variable_data = self._solution_variable_data()

        monitors_service = service_creator("monitors").create(
            fluent_connection._channel, fluent_connection._metadata, self._error_state
        )
        self.monitors = MonitorsManager(fluent_connection._id, monitors_service)
        self.events.register_callback(
            SolverEvent.SOLUTION_INITIALIZED, self.monitors.refresh
        )
        self.events.register_callback(SolverEvent.DATA_LOADED, self.monitors.refresh)

        fluent_connection.register_finalizer_cb(self.monitors.stop)

    def _solution_variable_data(self) -> SolutionVariableData:
        """Return the SolutionVariableData handle."""
        return service_creator("svar_data").create(
            self._solution_variable_service, self.fields.solution_variable_info
        )

    @property
    def svar_data(self):
        """``SolutionVariableData`` handle."""
        warnings.warn(
            "svar_data is deprecated. Use fields.solution_variable_data instead.",
            PyFluentDeprecationWarning,
        )
        return self.fields.solution_variable_data

    @property
    def svar_info(self):
        """``SolutionVariableInfo`` handle."""
        warnings.warn(
            "svar_info is deprecated. Use fields.solution_variable_info instead.",
            PyFluentDeprecationWarning,
        )
        return self.fields.solution_variable_info

    def _get_zones_info(self) -> list[ZoneInfo]:
        zones_info = []
        for (
            zone_info
        ) in self.fields.solution_variable_info.get_zones_info()._zones_info.values():
            zone_type = (
                ZoneType.CELL
                if zone_info.thread_type == SvarProtoModule.ThreadType.CELL_THREAD
                else ZoneType.FACE
            )
            zones_info.append(
                ZoneInfo(
                    _id=zone_info.zone_id, name=zone_info.name, zone_type=zone_type
                )
            )
        return zones_info

    @property
    def reduction(self):
        """``Reduction`` handle."""
        warnings.warn(
            "reduction is deprecated. Use fields.reduction instead.",
            PyFluentDeprecationWarning,
        )
        return self.fields.reduction

    @property
    def _version(self):
        """Fluent's product version."""
        if self._fluent_version is None:
            self._fluent_version = get_version_for_file_name(session=self)
        return self._fluent_version

    @property
    def tui(self):
        """Instance of ``main_menu`` on which Fluent's SolverTUI methods can be
        executed."""
        if self._tui is None:
            self._tui = _make_tui_module(self, "solver")

        return self._tui

    @property
    def workflow(self):
        """Datamodel root for workflow."""
        if not self._workflow:
            self._workflow = ClassicWorkflow(
                _make_datamodel_module(self, "workflow"),
                Solver,
                self.get_fluent_version(),
            )
        return self._workflow

    @classmethod
    def _interrupt(cls, command):
        interruptible_commands = [
            "solution/run-calculation/iterate",
            "solution/run-calculation/calculate",
            "solution/run-calculation/dual-time-iterate",
        ]
        if pyfluent.SUPPORT_SOLVER_INTERRUPT:
            if command.path in interruptible_commands:
                command._root.solution.run_calculation.interrupt()

    @property
    def settings(self):
        """Root settings object."""
        if self._settings_root is None:
            self._settings_root = flobject.get_root(
                flproxy=self._settings_service,
                version=self._version,
                interrupt=Solver._interrupt,
                file_transfer_service=self._file_transfer_service,
                scheme_eval=self.scheme_eval.scheme_eval,
            )
        return self._settings_root

    @property
    def system_coupling(self):
        """System coupling object."""
        if self._system_coupling is None:
            self._system_coupling = SystemCoupling(self)
        return self._system_coupling

    @property
    def preferences(self):
        """Datamodel root of preferences."""
        if self._preferences is None:
            self._preferences = _make_datamodel_module(self, "preferences")
        return self._preferences

    def _sync_from_future(self, fut: Future):
        with self._lck:
            try:
                fut_session = fut.result()
            except Exception as ex:
                raise RuntimeError("Unable to read mesh") from ex
            state = self.settings.get_state()
            super(Solver, self)._build_from_fluent_connection(
                fut_session._fluent_connection,
                fut_session._fluent_connection._connection_interface.scheme_eval,
                event_type=SolverEvent,
            )
            self._build_from_fluent_connection(
                fut_session._fluent_connection,
                fut_session._fluent_connection._connection_interface.scheme_eval,
            )
            # TODO temporary fix till set_state at settings root is fixed
            _set_state_safe(self.settings, state)

    def read_case_lightweight(self, file_name: str):
        """Read a case file using light IO mode.

        Parameters
        ----------
        file_name : str
            Case file name
        """
        import ansys.fluent.core as pyfluent

        self.file.read(file_type="case", file_name=file_name, lightweight_setup=True)
        launcher_args = dict(self._launcher_args)
        launcher_args.pop("lightweight_mode", None)
        launcher_args["case_file_name"] = file_name
        fut: Future = asynchronous(pyfluent.launch_fluent)(**launcher_args)
        fut.add_done_callback(functools.partial(Solver._sync_from_future, self))

    def get_state(self) -> StateT:
        """Get the state of the object."""
        return self.settings.get_state()

    def set_state(self, state: StateT | None = None, **kwargs):
        """Set the state of the object."""
        self.settings.set_state(state, **kwargs)

    def __call__(self):
        return self.get_state()

    def _populate_settings_api_root(self):
        if not self._settings_api_root:
            self._settings_api_root = _import_settings_root(self.settings)

    def __getattr__(self, attr):
        self._populate_settings_api_root()
        if attr in [x for x in dir(self._settings_api_root) if not x.startswith("_")]:
            if self.get_fluent_version() > FluentVersion.v242:
                warnings.warn(
                    f"'{attr}' is deprecated. Use 'settings.{attr}' instead.",
                    DeprecatedSettingWarning,
                )
        return getattr(self._settings_api_root, attr)

    def __dir__(self):
        settings_dir = []
        if self.get_fluent_version() <= FluentVersion.v242:
            self._populate_settings_api_root()
            settings_dir = dir(self._settings_api_root)
        dir_list = set(list(self.__dict__.keys()) + dir(type(self)) + settings_dir) - {
            "svar_data",
            "svar_info",
            "reduction",
            "field_data",
            "field_info",
            "field_data_streaming",
            "start_journal",
            "stop_journal",
        }
        return sorted(dir_list)



"""Sphinx documentation configuration file."""

from dataclasses import is_dataclass
from datetime import datetime
import inspect
import os
import platform
import subprocess

from ansys_sphinx_theme import ansys_favicon, get_version_match, pyansys_logo_black
from sphinx_gallery.sorting import FileNameSortKey

from ansys.fluent.core import __version__

# -- Project information -----------------------------------------------------

project = "ansys.fluent.core"
copyright = f"(c) {datetime.now().year} ANSYS, Inc. All rights reserved"
author = "ANSYS Inc."
cname = os.getenv("DOCUMENTATION_CNAME", "nocname.com")

# The short X.Y version
release = version = __version__

# -- General configuration ---------------------------------------------------
extensions = [
    "jupyter_sphinx",
    "notfound.extension",
    "numpydoc",
    "autodocsumm",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.coverage",
    "sphinx.ext.doctest",
    "sphinx.ext.extlinks",
    "sphinx.ext.intersphinx",
    "sphinx_autodoc_typehints",
    "sphinx_copybutton",
    "sphinxemoji.sphinxemoji",
    "sphinx_toggleprompt",
]

toggleprompt_offset_right = 35

if os.getenv("PYFLUENT_DOC_SKIP_EXAMPLES") != "1":
    extensions.append("sphinx_gallery.gen_gallery")

typehints_document_rtype = False

# Intersphinx mapping
intersphinx_mapping = {
    "python": ("https://docs.python.org/", None),
    "numpy": ("https://numpy.org/doc/stable", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable", None),
}

# SS01, SS02, SS03, GL08 all need clean up in the code to be reactivated.
# numpydoc configuration
numpydoc_use_plots = True
numpydoc_show_class_members = False
numpydoc_xref_param_type = True
numpydoc_validate = True
numpydoc_validation_checks = {
    "GL06",  # Found unknown section
    "GL07",  # Sections are in the wrong order.
    "GL08",  # The object does not have a docstring
    "GL09",  # Deprecation warning should precede extended summary
    "GL10",  # reST directives {directives} must be followed by two colons
    "SS01",  # No summary found
    "SS02",  # Summary does not start with a capital letter
    "SS03",  # Summary does not end with a period
    "SS04",  # Summary contains heading whitespaces
    # "SS05", # Summary must start with infinitive verb, not third person
    "RT02",  # The first line of the Returns section should contain only the
    # type, unless multiple values are being returned"
}
numpydoc_validation_exclude = {
    "ansys.fluent.core.generated.solver.settings_231.",
    "ansys.fluent.core.generated.solver.settings_232.",
    "ansys.fluent.core.generated.solver.settings_241.",
    "ansys.fluent.core.generated.solver.settings_242.",
    "ansys.fluent.core.generated.solver.settings_251.",
    "ansys.fluent.core.services.batch_ops.BatchOps.__init__",
}

# Favicon
html_favicon = ansys_favicon

# static path
html_static_path = ["_static"]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix(es) of source file names.
source_suffix = ".rst"

# The master toctree document.
master_doc = "index"

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = "en"

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = False

# Copy button customization ---------------------------------------------------
# exclude traditional Python prompts from the copied code
copybutton_prompt_text = r">>> ?|\.\.\. "
copybutton_prompt_is_regexp = True


def _stop_fluent_container(gallery_conf, fname):
    try:
        is_linux = platform.system() == "Linux"
        container_names = (
            subprocess.check_output(
                "docker container ls --format {{.Names}}", shell=is_linux
            )
            .decode("utf-8")
            .strip()
            .split()
        )
        for container_name in container_names:
            print(f"Container still running for script {fname}")
            subprocess.run(f"docker stop {container_name}", shell=is_linux)
    except Exception:
        pass


# -- Sphinx Gallery Options ---------------------------------------------------
sphinx_gallery_conf = {
    # convert rst to md for ipynb
    # "pypandoc": True,
    # path to your examples scripts
    "examples_dirs": ["../../examples/"],
    # path where to save gallery generated examples
    "gallery_dirs": ["examples"],
    # Pattern to search for example files
    "filename_pattern": r"exhaust_system_settings_api\.py|external_compressible_flow\.py|mixing_elbow_settings_api\.py|modeling_cavitation\.py|species_transport\.py|ahmed_body_workflow\.py|brake\.py|DOE_ML\.py|radiation_headlamp\.py|parametric_static_mixer_1\.py|conjugate_heat_transfer\.py|tyler_sofrin_modes\.py|lunar_lander_thermal\.py|modeling_ablation\.py",
    # Do not execute examples
    "plot_gallery": False,
    # Remove the "Download all examples" button from the top level gallery
    "download_all_examples": False,
    # Sort gallery example by file name instead of number of lines (default)
    "within_subsection_order": FileNameSortKey,
    # directory where function granular galleries are stored
    "backreferences_dir": None,
    # Modules for which function level galleries are created.  In
    "doc_module": "ansys-fluent-core",
    "thumbnail_size": (350, 350),
    "reset_modules_order": "after",
    "reset_modules": (_stop_fluent_container),
    "capture_repr": (),
    "remove_config_comments": True,
    "abort_on_example_error": True,
}


# -- Options for HTML output -------------------------------------------------
html_short_title = html_title = "PyFluent"
html_theme = "ansys_sphinx_theme"
html_logo = pyansys_logo_black
html_theme_options = {
    "github_url": "https://github.com/ansys/pyfluent",
    "show_prev_next": False,
    "show_breadcrumbs": True,
    "additional_breadcrumbs": [
        ("PyAnsys", "https://docs.pyansys.com/"),
    ],
    "icon_links": [
        {
            "name": "Support",
            "url": "https://github.com/ansys/pyfluent/discussions",
            "icon": "fa fa-comment fa-fw",
        },
    ],
    "switcher": {
        "json_url": f"https://{cname}/versions.json",
        "version_match": get_version_match(__version__),
    },
    "navbar_end": ["version-switcher", "theme-switcher", "navbar-icon-links"],
    "navigation_depth": -1,
    "collapse_navigation": True,
}

if os.getenv("PYFLUENT_DOC_SKIP_CHEATSHEET") != "1":
    html_theme_options["cheatsheet"] = {
        "file": "cheatsheet/cheat_sheet.qmd",
        "pages": ["index", "getting_started/index", "user_guide/index"],
        "title": "PyFluent cheat sheet",
        "version": __version__,
    }

# -- Options for HTMLHelp output ---------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = "pyfluentdoc"


# -- Options for LaTeX output ------------------------------------------------
latex_elements = {}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (
        master_doc,
        f"pyfluent-Documentation-{__version__}.tex",
        "ansys.fluent.core Documentation",
        author,
        "manual",
    ),
]


# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (
        master_doc,
        "ansys.fluent.core",
        "ansys.fluent.core Documentation",
        [author],
        1,
    )
]


# -- Options for Texinfo output ----------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        master_doc,
        "ansys.fluent.core",
        "ansys.fluent.core Documentation",
        author,
        "ansys.fluent.core",
        "Python client library that interfaces to Fluent",
        "Engineering Software",
    ),
]


# -- Options for Epub output -------------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ["search.html"]


def _get_cls(func):
    """Get the class containing the function."""
    module = inspect.getmodule(func)
    cls = module
    for name in func.__qualname__.split(".")[0:-1]:
        cls = getattr(cls, name)
    return cls


def skip(app, what, name, obj, would_skip, options):
    """Skip some specific members from documentation."""
    if name == "__init__":
        cls = _get_cls(obj)
        if is_dataclass(cls):
            # skip __init__ for dataclasses
            return True
    return would_skip


def setup(app):
    """Setup function for Sphinx builder."""
    app.connect("autodoc-skip-member", skip)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Classes for locally defining a parametric study for Fluent without running Fluent.
The study can then be submitted to be executed in parallel.

Example
-------

Set up a local study

>>> from ansys.fluent.core.parametric import LocalParametricStudy
>>> local_study = LocalParametricStudy(case_filepath="E:/elbow1_param.cas.h5")
>>> design_point = local_study.design_point("Base DP")
>>> design_point.input_parameters['v1'] = 0.0
>>> for idx in range(1, 20):
>>>   design_point = local_study.add_design_point("dp_"+str(idx))
>>>   design_point.input_parameters['v1'] = float(idx)/10.0

Run in Fluent

>>> local_study.run_in_fluent(5)

Display results

>>> for design_point in local_study.design_point_table:
>>>   for k, v in design_point.input_parameters.items():
>>>     print("input parameter", k, v)
>>>   for k, v in design_point.output_parameters.items():
>>>     print("output parameter", k, v)
>>> print(72 * "-")
"""

from math import ceil
from typing import Any, Dict

from ansys.fluent.core.launcher.launcher import launch_fluent
from ansys.fluent.core.utils.execution import asynchronous

BASE_DP_NAME = "Base DP"


def convert_design_point_parameter_units(
    value: Dict[str, float | int | str]
) -> Dict[str, float | int]:
    """Convert design point parameter units."""

    def conv(val):
        if type(val) in (float, int):
            return val
        if type(val) is not str:
            raise RuntimeError("Invalid value type for input parameter", val, type(val))
        pos = val.find(" [")
        if pos == -1:
            return float(val)
        return float(val[:pos])

    return {k: conv(v) for k, v in value.items()}


class LocalDesignPoint:
    """Purely local version of a design point in a parametric study.

    Parameters
    ----------
    name : str
        Name of the design point.
    output_parameters : dict
        Dictionary of output parameters (name of parameter to value).
    input_parameters : dict
        Dictionary of input parameters (name of parameter to value).
    status : DesignPointStatus
        Current status of the design point.
    """

    def __init__(self, design_point_name: str, base_design_point=None):
        """Initialize LocalDesignPoint."""
        self.name = design_point_name
        if base_design_point:
            self.__inputs = base_design_point.input_parameters.copy()
            self.__outputs = base_design_point.output_parameters.copy()
        else:
            self.__inputs = {}
            self.__outputs = {}

    @property
    def input_parameters(self) -> dict:
        """Get the input parameters."""
        return self.__inputs

    @input_parameters.setter
    def input_parameters(self, inputs: dict):
        self.__inputs = inputs

    @property
    def output_parameters(self) -> dict:
        """Get the output parameters."""
        return self.__outputs

    @output_parameters.setter
    def output_parameters(self, outputs: dict):
        self.__outputs = outputs


class LocalDesignPointTable(list):
    """Local version of the design point table in a parametric study.

    Methods
    -------
    add_design_point(design_point_name: str) -> DesignPoint
        Add a new design point to the table with the provided name.
    find_design_point(idx_or_name)
        Get a design point, either by name (str) or an index
        indicating the position in the table (by order of insertion).
    remove_design_point(idx_or_name)
        Remove a design point, either by name (str) or an index
        indicating the position in the table (by order of insertion).

    Raises
    ------
    RuntimeError
        If the design point is not found.
    """

    def __init__(self, base_design_point: LocalDesignPoint):
        """Initialize LocalDesignPointTable."""
        super().__init__()
        self.append(base_design_point)

    def add_design_point(self, design_point_name: str) -> LocalDesignPoint:
        """Add the design point."""
        self.append(LocalDesignPoint(design_point_name, self[0]))
        return self[-1]

    def find_design_point(self, idx_or_name) -> LocalDesignPoint:
        """Find the design point.

        Raises
        ------
        RuntimeError
            If the design point is not found.
        """
        if isinstance(idx_or_name, int):
            return self[idx_or_name]
        for design_point in self:
            if idx_or_name == design_point.name:
                return design_point
        raise RuntimeError(f"This design point is not found: {idx_or_name}")

    def remove_design_point(self, idx_or_name):
        """Remove the design point.

        Raises
        ------
        RuntimeError
            If the design point can not be removed.
        """
        design_point = self.find_design_point(idx_or_name)
        if design_point is self[0]:
            raise RuntimeError("You cannot remove the base design point.")
        self.remove(self.find_design_point(idx_or_name))


def _run_local_study_in_fluent(
    local_study,
    num_servers: int,
    launcher: Any,
    start_transcript: bool,
    capture_report_data: bool,
):
    source_table_size = len(local_study.design_point_table)

    def make_input_for_study(design_point_range) -> None:
        if design_point_range is None:
            design_point_range = range(0, source_table_size)
        study_input = []
        for idx in design_point_range:
            design_point = local_study.design_point(idx_or_name=idx)
            study_input.append(design_point.input_parameters.copy())
        return study_input

    def make_input_for_studies(num_servers) -> None:
        study_inputs = []
        total_num_points = num_points = source_table_size
        for i in range(num_servers):
            count = ceil(num_points / num_servers)
            range_base = total_num_points - num_points
            num_points -= count
            num_servers -= 1
            study_inputs.append(
                make_input_for_study(range(range_base, range_base + count))
            )
        return study_inputs

    @asynchronous
    def make_parametric_session(case_filepath):
        return launcher(case_file_name=case_filepath, start_transcript=start_transcript)

    @asynchronous
    def apply_to_study(study, inputs):
        for input in inputs:
            dp_names = set([*study.design_points.keys()])
            try:
                study.design_points.create_1()
            except AttributeError:
                study.design_points.create()
            dp1_name = set([*study.design_points.keys()]).difference(dp_names).pop()
            dp = study.design_points[dp1_name]
            dp.capture_simulation_report_data = capture_report_data
            dp.input_parameters = convert_design_point_parameter_units(input.copy())

    @asynchronous
    def update_design_point(study):
        study.design_points.update_all()

    def apply_to_studies(studies, inputs) -> None:
        results = []
        for item in list(zip(studies, inputs)):
            study, input = item
            results.append(apply_to_study(study, input))
        for result in results:
            result.result()

    study_inputs = make_input_for_studies(num_servers)

    sessions = []
    studies = []
    for i in range(num_servers):
        sessions.append(
            make_parametric_session(case_filepath=local_study.case_filepath)
        )

    for session in sessions:
        session.result().settings.parametric_studies.initialize(
            project_filename=session.result().id
        )
        for study in session.result().settings.parametric_studies.values():
            studies.append(study)

    apply_to_studies(studies, study_inputs)

    updates = []
    for study in studies:
        updates.append(update_design_point(study))

    for update in updates:
        update.result()

    for study in studies:
        for _, design_point in study.design_points.items():
            for local_design_point in local_study.design_point_table:
                local_design_point.output_parameters = (
                    design_point.output_parameters.get_state()
                )


class LocalParametricStudy:
    """Local version of a parametric study that manages design points to parametrize a
    Fluent solver setup.

    Methods
    -------
    add_design_point(design_point_name)
        Add a design point.
    design_point(idx_or_name)
        Get a design point, either by name (str) or an index
        indicating the position in the table (by order of insertion).
    run_in_fluent()
        Run the study in Fluent

    Raises
    ------
    RuntimeError
        If the design point is not found.
    """

    def __init__(self, case_filepath: str, base_design_point_name: str = "Base DP"):
        """Initialize LocalParametricStudy."""
        from ansys.fluent.core.filereader.casereader import CaseReader

        self.case_filepath = case_filepath
        base_design_point = LocalDesignPoint(base_design_point_name)
        case_reader = CaseReader(case_file_name=case_filepath)

        base_design_point.input_parameters = {
            p.name: p.value for p in case_reader.input_parameters()
        }

        base_design_point.output_parameters = {
            p.name: None for p in case_reader.output_parameters()
        }

        self.design_point_table = LocalDesignPointTable(base_design_point)

    def add_design_point(self, design_point_name: str) -> LocalDesignPoint:
        """Add the design point."""
        return self.design_point_table.add_design_point(design_point_name)

    def design_point(self, idx_or_name) -> LocalDesignPoint:
        """Get the design point."""
        return self.design_point_table.find_design_point(idx_or_name)

    def run_in_fluent(
        self,
        num_servers: int,
        launcher: Any = launch_fluent,
        start_transcript: bool = False,
        capture_report_data: bool = False,
    ):
        """Run the local study in fluent."""
        _run_local_study_in_fluent(
            local_study=self,
            num_servers=num_servers,
            launcher=launcher,
            start_transcript=start_transcript,
            capture_report_data=capture_report_data,
        )



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing class encapsulating Fluent connection.

**********PRESENTLY SAME AS SOLVER WITH A SWITCH TO SOLVER***********
"""

from typing import Any, Dict

from ansys.fluent.core.session_solver import Solver


class SolverLite(Solver):
    """Encapsulates a Fluent - Solver(Lite) session connection.
    SolverLite(Session) holds the top-level objects
    for solver TUI and settings objects calls."""

    def __init__(
        self,
        fluent_connection=None,
        scheme_eval=None,
        start_transcript: bool = True,
        launcher_args: Dict[str, Any] | None = None,
    ):
        """SolverLite session.

        Parameters
        ----------
            fluent_connection (:ref:`ref_fluent_connection`): Encapsulates a Fluent connection.
            scheme_eval: SchemeEval
                Instance of ``SchemeEval`` to execute Fluent's scheme code on.
            start_transcript : bool, optional
                Whether to start the Fluent transcript in the client.
                The default is ``True``, in which case the Fluent transcript can be subsequently
                started and stopped using method calls on the ``Session`` object.
        """
        super().__init__(
            fluent_connection=fluent_connection,
            scheme_eval=scheme_eval,
            start_transcript=start_transcript,
            launcher_args=launcher_args,
        )
        self._tui_service = self._datamodel_service_tui
        self._settings_service = self.settings_service
        self._tui = None
        self._settings_root = None

    # One can inherit methods from 'Solver' and re-define it here to make it unavailable in solver-Lite

    def switch_to_full_solver(self):
        """A switch to move to the full-solver session from solver-lite."""
        solver_session = Solver(
            fluent_connection=self._fluent_connection, scheme_eval=self.scheme_eval
        )
        return solver_session



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


def assign_task_arguments(
    workflow, check_state: bool, task_name: str, **kwargs
) -> None:
    task = workflow.TaskObject[task_name]
    task.Arguments = kwargs
    if check_state:
        # the state that we have set must be a subset of the total state
        assert kwargs.items() <= task.Arguments().items()


def check_task_execute_preconditions(task) -> None:
    assert task.State() == "Out-of-date"
    assert not task.Errors() or not len(task.Errors())


def check_task_execute_postconditions(task) -> None:
    assert task.State() == "Up-to-date"
    assert not task.Errors() or not len(task.Errors())


def execute_task_with_pre_and_postcondition_checks(workflow, task_name: str) -> None:
    task = workflow.TaskObject[task_name]
    check_task_execute_preconditions(task)
    # Some tasks are wrongly returning False in meshing workflow itself
    # so we add a temporary caveat below
    result = task.Execute()
    if task_name not in (
        "Add Local Sizing",
        "Add Boundary Layers",
        "Import CAD and Part Management",
    ):
        assert result is True
    check_task_execute_postconditions(task)



# Copyright (C) 2021 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Test file transfer service."""

import os
import pathlib

import pytest

from ansys.fluent.core import examples
from ansys.fluent.core.utils.file_transfer_service import (
    LocalFileTransferStrategy,
    RemoteFileTransferStrategy,
)


def file_downloaded_to_the_client(file_name: str) -> bool:
    """Check if client file exists.

    Parameters
    ----------
    file_name: str
        File name.

    Returns
    -------
        Whether file exists.
    """
    full_file_name = pathlib.Path(
        "/home/runner/.local/share/ansys_fluent_core/examples"
    ) / os.path.basename(file_name)
    return full_file_name.is_file()


@pytest.mark.codegen_required
@pytest.mark.fluent_version(">=24.2")
def test_remote_grpc_fts_container(
    monkeypatch, new_solver_session, new_meshing_session
):
    solver = new_solver_session
    import_case_file_name = examples.download_file(
        "mixing_elbow.cas.h5", "pyfluent/mixing_elbow"
    )
    import_mesh_file_name = examples.download_file(
        "mixing_elbow.msh.h5", "pyfluent/mixing_elbow"
    )
    solver.file.read_case(file_name=import_case_file_name)
    if solver._file_transfer_service:
        solver.file.write_case(file_name="downloaded_solver_mixing_elbow.cas.h5")
        assert solver.file_exists_on_remote("downloaded_solver_mixing_elbow.cas.h5")
        assert file_downloaded_to_the_client("downloaded_solver_mixing_elbow.cas.h5")

    meshing = new_meshing_session
    meshing.meshing.File.ReadMesh(FileName=import_mesh_file_name)
    if meshing._file_transfer_service:
        meshing.meshing.File.WriteMesh(
            FileName="downloaded_meshing_mixing_elbow.msh.h5"
        )
        assert meshing.file_exists_on_remote("downloaded_meshing_mixing_elbow.msh.h5")
        assert file_downloaded_to_the_client("downloaded_meshing_mixing_elbow.msh.h5")


@pytest.mark.standalone
def test_read_case_and_data(monkeypatch):
    import ansys.fluent.core as pyfluent

    monkeypatch.setattr(pyfluent, "USE_FILE_TRANSFER_SERVICE", True)

    case_file_name = examples.download_file(
        "mixing_elbow.cas.h5", "pyfluent/mixing_elbow"
    )
    data_file_name = examples.download_file(
        "mixing_elbow.dat.h5", "pyfluent/mixing_elbow"
    )
    assert case_file_name
    assert data_file_name
    solver = pyfluent.launch_fluent(file_transfer_service=LocalFileTransferStrategy())

    solver.file.read(file_type="case-data", file_name=case_file_name)
    solver.file.write(file_type="case-data", file_name="write_data.cas.h5")

    solver.file.read_case_data(file_name=case_file_name)
    solver.file.write_case_data(file_name="write_case_data.cas.h5")
    solver.exit()

    solver = pyfluent.launch_fluent(
        file_transfer_service=LocalFileTransferStrategy(server_cwd=os.getcwd())
    )

    solver.file.read(file_type="case-data", file_name=case_file_name)
    solver.file.write(file_type="case-data", file_name="write_data.cas.h5")

    solver.file.read_case_data(file_name=case_file_name)
    solver.file.write_case_data(file_name="write_case_data.cas.h5")
    solver.exit()


@pytest.mark.skip(reason="Skips upload even after adding ImportGeometry task object.")
def test_datamodel_execute():
    import ansys.fluent.core as pyfluent

    meshing = pyfluent.launch_fluent(
        mode="meshing", file_transfer_service=RemoteFileTransferStrategy()
    )
    meshing.workflow.InitializeWorkflow(WorkflowType="Watertight Geometry")
    import_geom = meshing.workflow.TaskObject["Import Geometry"]
    import_geom.Arguments = {"FileName": "geom"}

    with pytest.raises(RuntimeError):
        import_geom.Execute()


def test_file_list_in_datamodel(fault_tolerant_workflow_session):
    meshing = fault_tolerant_workflow_session
    fmd_file = examples.download_file("exhaust_system.fmd", "pyfluent/exhaust_system")
    meshing.PartManagement.AppendFmdFiles(
        AssemblyParentNode=0,
        FilePath=[fmd_file],
        FileUnit="mm",
        IgnoreSolidNamesAppend=False,
        JtLOD="1",
        Options={
            "Line": False,
            "Solid": True,
            "Surface": True,
        },
        PartPerBody=False,
        PrefixParentName=False,
        RemoveEmptyParts=True,
        Route="Native",
    )


@pytest.mark.standalone
def test_local_file_transfer_in_datamodel(monkeypatch):
    import ansys.fluent.core as pyfluent

    monkeypatch.setattr(pyfluent, "USE_FILE_TRANSFER_SERVICE", True)

    fmd_file = examples.download_file("exhaust_system.fmd", "pyfluent/exhaust_system")

    assert fmd_file

    meshing = pyfluent.launch_fluent(
        mode=pyfluent.FluentMode.MESHING,
        file_transfer_service=LocalFileTransferStrategy(),
    )

    meshing.workflow.InitializeWorkflow(WorkflowType="Fault-tolerant Meshing")

    meshing.PartManagement.AppendFmdFiles(
        AssemblyParentNode=0,
        FilePath=[fmd_file],
        FileUnit="mm",
        IgnoreSolidNamesAppend=False,
        JtLOD="1",
        Options={
            "Line": False,
            "Solid": True,
            "Surface": True,
        },
        PartPerBody=False,
        PrefixParentName=False,
        RemoveEmptyParts=True,
        Route="Native",
    )
    meshing.exit()
